"""Pytest translation of test_decimation_methods.py

This is a test to confirm that mth5's decimation method returns the same
default values as aurora's prototype decimate.
"""

import numpy as np
from mth5.helpers import close_open_files
from mth5.mth5 import MTH5
from mth5.processing import KernelDataset, RunSummary

from aurora.pipelines.time_series_helpers import prototype_decimate
from aurora.test_utils.synthetic.make_processing_configs import create_test_run_config


def test_decimation_methods_agree(worker_safe_test1_h5, synthetic_test_paths):
    """Test that aurora and mth5 decimation methods produce identical results."""
    close_open_files()
    mth5_path = worker_safe_test1_h5

    run_summary = RunSummary()
    run_summary.from_mth5s([mth5_path])
    tfk_dataset = KernelDataset()
    station_id = "test1"
    run_id = "001"
    tfk_dataset.from_run_summary(run_summary, station_id)

    processing_config = create_test_run_config(station_id, tfk_dataset)

    mth5_obj = MTH5(file_version="0.1.0")
    mth5_obj.open_mth5(mth5_path, mode="a")
    decimated_ts = {}

    for dec_level_id, dec_config in enumerate(processing_config.decimations):
        decimated_ts[dec_level_id] = {}
        if dec_level_id == 0:
            run_obj = mth5_obj.get_run(station_id, run_id, survey=None)
            run_ts = run_obj.to_runts(start=None, end=None)
            run_xrds = run_ts.dataset
            decimated_ts[dec_level_id]["run_xrds"] = run_xrds
            current_sample_rate = run_obj.metadata.sample_rate

        if dec_level_id > 0:
            run_xrds = decimated_ts[dec_level_id - 1]["run_xrds"]
            target_sample_rate = current_sample_rate / (dec_config.decimation.factor)

            decimated_1 = prototype_decimate(dec_config.decimation, run_xrds)
            decimated_2 = run_xrds.sps_filters.decimate(
                target_sample_rate=target_sample_rate
            )

            difference = decimated_2 - decimated_1
            assert np.isclose(difference.to_array(), 0).all()

            decimated_ts[dec_level_id]["run_xrds"] = decimated_1
            current_sample_rate = target_sample_rate



"""Pytest translation of test_make_h5s.py"""

from loguru import logger
from mth5.data.make_mth5_from_asc import create_test4_h5

from aurora.test_utils.synthetic.paths import _get_mth5_ascii_data_path


def test_get_mth5_ascii_data_path():
    """Make sure that the ascii data are where we think they are."""
    mth5_data_path = _get_mth5_ascii_data_path()
    ascii_file_paths = list(mth5_data_path.glob("*asc"))
    file_names = [x.name for x in ascii_file_paths]
    logger.info(f"mth5_data_path = {mth5_data_path}")
    logger.info(f"file_names = {file_names}")

    assert "test1.asc" in file_names
    assert "test2.asc" in file_names


def test_make_upsampled_mth5(synthetic_test_paths):
    """Test creating upsampled mth5 file using synthetic_test_paths fixture."""
    file_version = "0.2.0"
    create_test4_h5(
        file_version=file_version, source_folder=synthetic_test_paths.ascii_data_path
    )



"""
This module contains methods for generating processing config objects that are
used in aurora's tests of processing synthetic data.
"""

from typing import Optional, Union

from loguru import logger
from mth5.processing import KernelDataset, RunSummary

from aurora.config import BANDS_256_26_FILE, BANDS_DEFAULT_FILE
from aurora.config.config_creator import ConfigCreator
from aurora.test_utils.synthetic.paths import SyntheticTestPaths


synthetic_test_paths = SyntheticTestPaths()
CONFIG_PATH = synthetic_test_paths.config_path
MTH5_PATH = synthetic_test_paths.mth5_path
SUPPORTED_TEST_CASE_IDS = [
    "test1",
    "test2",
    "test1r2",
    "test2r1",
    "test1_tfk",
    "test1r2_tfk",
]


def create_test_run_config(
    test_case_id,
    kernel_dataset,
    matlab_or_fortran="",
    save="json",
    channel_nomenclature="default",
):
    """
    Use config creator to generate a processing config file for the synthetic data.

    Parameters
    ----------
    test_case_id: string
        Must be in SUPPORTED_TEST_CASE_IDS
    kernel_dataset: aurora.transfer_function.kernel_dataset.KernelDataset
        Description of the dataset to process
    matlab_or_fortran: str
        "", "matlab", "fortran"
    save: str
        if this has the value "json" a copy of the processing config will be written to a json file
        The json file name is p.json_fn() with p the processing config object.
    channel_nomenclature: str
        A label for the channel nomenclature.  This should be one of the values in
        mt_metadata/transfer_functions/processing/aurora/standards/channel_nomenclatures.json
        currently ["default", "lemi12", "lemi34", "phoenix123", "musgraves",]

    Returns
    -------
    p: aurora.config.metadata.processing.Processing
        The processing config object


    """
    estimation_engine = "RME"
    local_station_id = test_case_id
    remote_station_id = ""

    if test_case_id in ["test1r2", "test1r2_tfk"]:
        estimation_engine = "RME_RR"
        local_station_id = "test1"
        remote_station_id = "test2"
    if test_case_id == "test2r1":
        estimation_engine = "RME_RR"
        local_station_id = "test2"
        remote_station_id = "test1"

    if matlab_or_fortran == "matlab":
        emtf_band_setup_file = BANDS_256_26_FILE
        num_samples_window = 256
        num_samples_overlap = 64
        config_id = f"{local_station_id}-{matlab_or_fortran}"
    elif matlab_or_fortran == "fortran":
        emtf_band_setup_file = BANDS_DEFAULT_FILE
        num_samples_window = 128
        num_samples_overlap = 32
        config_id = f"{local_station_id}-{matlab_or_fortran}"
    else:
        emtf_band_setup_file = BANDS_DEFAULT_FILE
        num_samples_window = 128
        num_samples_overlap = 32
        config_id = f"{local_station_id}"

    cc = ConfigCreator()

    if test_case_id in ["test1", "test2"]:
        p = cc.create_from_kernel_dataset(
            kernel_dataset,
            emtf_band_file=emtf_band_setup_file,
            num_samples_window=num_samples_window,
        )
        p.id = config_id
        p.channel_nomenclature.keyword = channel_nomenclature
        p.set_default_input_output_channels()
        p.drop_reference_channels()

    elif test_case_id in ["test2r1", "test1r2"]:
        config_id = f"{config_id}-RR{remote_station_id}"
        p = cc.create_from_kernel_dataset(
            kernel_dataset,
            emtf_band_file=emtf_band_setup_file,
            num_samples_window=num_samples_window,
        )
        p.id = config_id
        p.channel_nomenclature.keyword = channel_nomenclature
        p.set_default_input_output_channels()
        p.set_default_reference_channels()

    elif test_case_id in ["test1_tfk", "test1r2_tfk"]:
        from aurora.general_helper_functions import BAND_SETUP_PATH

        emtf_band_setup_file = BAND_SETUP_PATH.joinpath("bs_six_level.cfg")
        if test_case_id == "test1r2_tfk":
            config_id = f"{config_id}-RR{remote_station_id}_tfk"
        p = cc.create_from_kernel_dataset(
            kernel_dataset,
            emtf_band_file=emtf_band_setup_file,
            num_samples_window=num_samples_window,
        )
        p.id = config_id
        p.channel_nomenclature.keyword = channel_nomenclature
        p.set_default_input_output_channels()
        p.set_default_reference_channels()

    for decimation in p.decimations:
        decimation.estimator.engine = estimation_engine
        decimation.stft.window.type = "hamming"
        decimation.stft.window.num_samples = num_samples_window
        decimation.stft.window.overlap = num_samples_overlap
        decimation.regression.max_redescending_iterations = 2
        if test_case_id == "test2":
            decimation.stft.window.type = "boxcar"

    if save == "json":
        CONFIG_PATH.mkdir(parents=True, exist_ok=True)
        filename = CONFIG_PATH.joinpath(p.json_fn())
        p.save_as_json(filename=filename)

    return p


def make_processing_config_and_kernel_dataset(
    config_keyword: str,
    station_id: str,
    remote_id: Optional[str] = None,  # TODO: allow empty str instead of None
    mth5s: Optional[Union[list, tuple]] = None,
    channel_nomenclature: Optional[str] = "default",
):
    """
    Gets the processing config and the tfk_dataset

    TODO: Move this to aurora/test_utils/synthetic/
     - this can then be used by test_fourier_coefficients to validate that the FCs are there before processing

    Parameters
    ----------
    config_keyword: str
        Must be in SUPPORTED_TEST_CASE_IDS
    station_id: str
        The name of the station to proces
    remote_id: str
    mth5s: Union[list, tuple]
    channel_nomenclature: Optional[str]

    Returns
    -------

    """
    run_summary = RunSummary()
    run_summary.from_mth5s(mth5s)

    # next two lines purely for codecov
    run_summary.print_mini_summary
    run_summary_clone = run_summary.clone()

    # run_summary.drop_runs_shorter_than(100000)
    tfk_dataset = KernelDataset()
    tfk_dataset.from_run_summary(
        run_summary_clone, local_station_id=station_id, remote_station_id=remote_id
    )
    processing_config = create_test_run_config(
        config_keyword, tfk_dataset, channel_nomenclature=channel_nomenclature
    )
    return tfk_dataset, processing_config


def test_to_from_json():
    """
    Intended to test that processing config can be stored as a json, then
    reloaded from json and is equal.

    WORK IN PROGRESS -- see mt_metadata Issue #222

    Development Notes:
    TODO: This test should be completed and moved into tests.
    The json does not load into an mt_metadata object.
    The problem seems to be that at the run-level of the processing config there is an
    intention to allow for multiple time-periods. This is reasonable, consider a station
    running for several months, we may want to only process data from certain chunks
    of the time series.
    However, the time period reader does not seem to work as expected.
    A partial fix is on fix_issue_222 branch of mt_metadata

    Related to issue #172


    Returns
    -------

    """
    # import pandas as pd
    from mt_metadata.processing.aurora import Processing
    from mth5.processing import KernelDataset, RunSummary

    # Specify path to mth5
    data_path = MTH5_PATH.joinpath("test1.h5")
    if not data_path.exists():
        logger.error("You need to run make_mth5_from_asc.py")
        raise Exception
    mth5_paths = [
        data_path,
    ]
    run_summary = RunSummary()
    run_summary.from_mth5s(mth5_paths)
    tfk_dataset = KernelDataset()
    station_id = "test1"
    tfk_dataset.from_run_summary(run_summary, station_id)

    processing_config = create_test_run_config(station_id, tfk_dataset, save="json")
    p = Processing()
    json_fn = CONFIG_PATH.joinpath(processing_config.json_fn())
    p.from_json(json_fn)
    logger.info("Assert equal needed here")
    # This fails (July 22, 2024)
    # assert p == processing_config

    # This should be true, but its false
    # p.stations.local.runs == processing_config.stations.local.runs
    # p.stations.local.runs[0] == processing_config.stations.local.runs[0]

    """
    Debugging Notes:
    Once the updated parsing from #222 is applied, the next problem is that the object that was
    read-back from json has two dicts in its time periods:
    p.stations.local.runs[0].time_periods
    [{'time_period': {'end': '1980-01-01T11:06:39+00:00', 'start': '1980-01-01T00:00:00+00:00'}},
     {'time_period': {'end': '1980-01-01T11:06:39+00:00', 'start': '1980-01-01T00:00:00+00:00'}}]
    processing_config.stations.local.runs[0].time_periods
    [{
        "time_period": {
            "end": "1980-01-01T11:06:39+00:00",
            "start": "1980-01-01T00:00:00+00:00"
        }
    }]
    """
    return


def main():
    """Allow the module to be called from the command line"""
    # TODO: fix test_to_from_json and put in tests.
    #  - see issue #222 in mt_metadata.
    test_to_from_json()
    # create_test_run_config("test1", df)
    # create_test_run_config("test2")
    # create_test_run_config("test1r2")
    # create_test_run_config("test2r1")


if __name__ == "__main__":
    main()



"""
Helper functions to handle workarounds.
"""

import numpy as np
from mt_metadata.transfer_functions import TF


def tfs_nearly_equal(tf1: TF, tf2: TF) -> bool:
    """
    Addresses issue # 381.
    Werid bug on MacOS involving numeric noise.

    Also forces the creation time of tf2 to match tf1.

    """
    cond1 = np.isclose(
        tf2.residual_covariance, tf1.residual_covariance, atol=1e-26, rtol=1e-30
    ).all()
    cond2 = np.isclose(
        tf2.transfer_function_error, tf1.transfer_function_error, atol=1e-14, rtol=1e-15
    ).all()
    cond3 = np.isclose(
        tf2.transfer_function, tf1.transfer_function, atol=1e-14, rtol=1e-15
    ).all()

    if cond1 and cond2 and cond3:
        tf2_copy = tf2.copy()
        tf2_copy.residual_covariance = tf1.residual_covariance
        tf2_copy.transfer_function_error = tf1.transfer_function_error
        tf2_copy.transfer_function = tf1.transfer_function
        # Triage the creation time
        tf2_copy.station_metadata.provenance.creation_time = (
            tf1.station_metadata.provenance.creation_time
        )
        # Triage the processed_date
        tf2_copy.station_metadata.transfer_function.processed_date = (
            tf1.station_metadata.transfer_function.processed_date
        )
        return tf1 == tf2_copy

    else:
        return False



"""
This module provides a TransferFunctionCollection container to hold:
1. TransferFunctionHeader
2. Dictionary of TransferFunction Objects

Development Notes:
 This class could use substantial updating.  There are two main issues to be aware of.

 First, the transfer function objects that is "collects" are themselves things
 that could use substantial updates.  This is because the aurora TransferFunction
 objects to not use the mt_metadata.transfer_function.TF and they should.  This
 is issue #203.

 Second, the Collection can have more than one value for a TF at a given
 frequency or period because the TFs can be estimated from different
 decimation levels.

 A single transfer function object is associated with a station, here called
 the "local_station". In a database of TFs could add a column for
 local_station and one for reference station.
"""

import pathlib
from typing import Any, Optional, Union

import numpy as np
import xarray as xr
from loguru import logger
from mt_metadata.processing.aurora.channel_nomenclature import ChannelNomenclature

from aurora.config.metadata.processing import Processing
from aurora.general_helper_functions import FIGURES_PATH


EMTF_REGRESSION_ENGINE_LABELS = {}
EMTF_REGRESSION_ENGINE_LABELS["RME"] = "Robust Single Station"
EMTF_REGRESSION_ENGINE_LABELS["RME_RR"] = "Robust Remote Reference"


def plot_rho(
    ax,
    periods,
    rho,
    marker="o",
    color="k",
    linestyle="None",
    label="",
    markersize=10,
    ax_label_size=16,
):
    """

    Plots apparent resistivity on the given axis

    Parameters
    ----------
    ax
    periods
    rho
    marker
    color
    linestyle
    label
    markersize
    ax_label_size

    Returns
    -------

    """
    ax.loglog(
        periods,
        rho,
        marker=marker,
        color=color,
        linestyle=linestyle,
        label=label,
        markersize=markersize,
    )
    ax.tick_params(axis="both", which="major", labelsize=ax_label_size)
    ax.tick_params(axis="x", which="minor", bottom=True)
    return


def plot_phi(
    ax,
    periods,
    phi,
    marker="o",
    color="k",
    linestyle="None",
    label="",
    markersize=10,
    ax_label_size=16,
):
    """
    Plots the phase on the given axis.

    Parameters
    ----------
    ax
    periods
    phi
    marker
    color
    linestyle
    label
    markersize
    ax_label_size

    Returns
    -------

    """
    ax.semilogx(
        periods,
        phi,
        marker=marker,
        color=color,
        linestyle=linestyle,
        label=label,
        markersize=markersize,
    )
    ax.tick_params(axis="both", which="major", labelsize=ax_label_size)
    ax.minorticks_on()  # (axis="x", which="minor", bottom=True)
    return


class TransferFunctionCollection(object):
    def __init__(
        self,
        tf_dict: Optional[Union[dict, None]] = None,
        processing_config: Optional[Union[Processing, None]] = None,
    ):
        """
        Constructor.

        Parameters
        ----------
        kwargs

        tf_dict: dict
            This is a dictionary of TTFZ objects, one per decimation level.  They are
            keyed by the decimation_level_id, usually integers 0, 1, 2... n_dec
        processing_config: Processing
            The processing parameters associated with the TF objects in the dict.

        """
        self.tf_dict = tf_dict
        self.processing_config = processing_config
        self.merged_tf = None
        self.merged_cov_nn = None
        self.merged_cov_ss_inv = None

    @property
    def header(self) -> dict:
        """returns the tf_header (ordered dict) from the first tf in the dict"""
        return self.tf_dict[0].tf_header

    @property
    def local_station_id(self) -> str:
        """return the local station id"""
        return self.header.local_station.id

    @property
    def remote_station_id(self) -> str:
        """return the remote station id"""
        if self.header.remote_station:
            return self.header.remote_station[0].id
        else:
            return ""

    @property
    def total_number_of_frequencies(self) -> int:
        """
        Returns int:
            Count of number of frequencies over all TFs in dict
        """
        num_frequecies = 0
        for dec_level in self.tf_dict.keys():
            num_frequecies += len(self.tf_dict[dec_level].periods)
        return num_frequecies

    @property
    def channel_list(self) -> list:
        """
        Gets all channel names.
         - assumes the first TF in dict is representative of all

        Returns
        -------
        all_channels: list
            All the channel names for the first TF
        """
        tf_input_chs = self.tf_dict[0].transfer_function.input_channel.data.tolist()
        tf_output_chs = self.tf_dict[0].transfer_function.output_channel.data.tolist()
        all_channels = tf_input_chs + tf_output_chs
        return all_channels

    @property
    def total_number_of_channels(self) -> int:
        """Get a count of all channels (input and output) in the TF"""
        num_channels = 0
        num_channels += len(self.header.input_channels)
        num_channels += len(self.header.output_channels)
        return num_channels

    @property
    def number_of_decimation_levels(self) -> int:
        """
        Returns a count of number of decimation levels
        - assumes that each entry in dict is a separate decimation level.
        """
        return len(self.tf_dict)

    def get_merged_dict(self, channel_nomenclature) -> dict:
        """
        merges all decimation levels and returns a dict

        Returns
        -------
        output: dict
            merged tf values
            - this can be used to assign values to an mt_metadata.transfer_function.TF
        """
        output = {}
        self._merge_decimation_levels()
        self.check_all_channels_present(channel_nomenclature)
        output["tf"] = self.merged_tf
        output["cov_ss_inv"] = self.merged_cov_ss_inv
        output["cov_nn"] = self.merged_cov_nn
        return output

    def _merge_decimation_levels(self) -> None:
        """
        Merge all decimation levels into a single xarray for export.

        Development Notes:
         The output of this may become its own class, MergedTransferFunction

         One concern here is that the same period can be estimted at more then one
         decimation level, making the frequency or period axis not of the same order
         as the number of estimates.  Ways around this:
         1. We can multi-index the period axis with decimation level
         2. Make sure that the processing config does not dupliacate periods
         3. Take the average estimate over all periods heving the same value
         4. Drop all but one estimate accoring to a rule (keep most observations say)

         Does TFXML support multiple estimates? How is it dealt with there?
         There is an issue here -

        Flow:
        starting from self.tf_dict, cast each decimation level's tf, which is
        an xr.DataArray to xr.Dataset, forming a list of datasets.
        This dataset list is then combined over all periods forming a representation
        of the TTFZ which is merged over all the decimation levels.

        result is to build:
        merged_tf
        merged_cov_ss_inv
        merged_cov_nn
        """
        n_dec = self.number_of_decimation_levels

        tmp = [self.tf_dict[i].tf.to_dataset("period") for i in range(n_dec)]
        merged_tf = xr.combine_by_coords(tmp)
        merged_tf = merged_tf.to_array(dim="period")
        self.merged_tf = merged_tf

        tmp = [self.tf_dict[i].cov_ss_inv.to_dataset("period") for i in range(n_dec)]
        merged_cov_ss_inv = xr.combine_by_coords(tmp)
        merged_cov_ss_inv = merged_cov_ss_inv.to_array("period")
        self.merged_cov_ss_inv = merged_cov_ss_inv

        tmp = [self.tf_dict[i].cov_nn.to_dataset("period") for i in range(n_dec)]
        merged_cov_nn = xr.combine_by_coords(tmp)
        merged_cov_nn = merged_cov_nn.to_array("period")
        self.merged_cov_nn = merged_cov_nn

        return

    def check_all_channels_present(
        self, channel_nomenclature: ChannelNomenclature
    ) -> None:
        """
        Checks if TF has tipper.  If not, fill in the tipper data with NaN and also
         update the noise covariance matrix so shape is as expected by mt_metadata.

        Development Notes:
            For testing, at least some Parkfield tests have no hz.
            Synthetic tests do not (currently) have missing hz.

        Parameters
        ----------
        channel_nomenclature: ChannelNomenclature
            Scheme according to how channels are named

        """
        ex, ey, hx, hy, hz = channel_nomenclature.unpack()
        if hz not in self.merged_tf.output_channel.data.tolist():
            output_channels_original = self.merged_tf.output_channel.data.tolist()
            output_channels = [f"tmp___{x}" for x in output_channels_original]
            output_channels[0] = hz
            tmp = self.merged_tf.copy(deep=True)
            tmp = tmp.assign_coords({"output_channel": output_channels})
            tmp.data *= np.nan
            tmp = tmp.to_dataset("period")
            tmp = tmp.merge(self.merged_tf.to_dataset("period"))
            tmp = tmp.to_array("period")
            output_channels = tmp.output_channel.data.tolist()
            output_channels = [x for x in output_channels if x[0:6] != "tmp___"]
            tmp = tmp.sel(output_channel=output_channels)
            self.merged_tf = tmp

            n_output_ch = len(self.merged_tf.output_channel)  # this should be 3
            n_periods = len(self.merged_tf.period)
            cov_nn_dims = (n_output_ch, n_output_ch, n_periods)
            cov_nn_data = np.zeros(cov_nn_dims, dtype=np.complex128)

            cov_nn = xr.DataArray(
                cov_nn_data,
                dims=["output_channel_1", "output_channel_2", "period"],
                coords={
                    "output_channel_1": self.merged_tf.output_channel.data,
                    "output_channel_2": self.merged_tf.output_channel.data,
                    "period": self.merged_tf.period,
                },
            )

            # to dataset and back makes the coords line up -- this does not seem robust
            cov_nn = cov_nn.to_dataset("period").to_array("period")

            for out_ch_1 in cov_nn.output_channel_1.data.tolist():
                for out_ch_2 in cov_nn.output_channel_1.data.tolist():
                    try:
                        values = self.merged_cov_nn.loc[:, out_ch_1, out_ch_2]
                        cov_nn.loc[:, out_ch_1, out_ch_2] = values
                    except KeyError:
                        pass
            self.merged_cov_nn = cov_nn
        if ex not in self.merged_tf.output_channel.data.tolist():
            # placeholder: logic for missing ex would go here
            msg = "No handling for missing ex channel"
            raise NotImplementedError(msg)
        if ey not in self.merged_tf.output_channel.data.tolist():
            # placeholder: logic for missing ey would go here
            msg = "No handling for missing ey channel"
            raise NotImplementedError(msg)

    def rho_phi_plot(
        self,
        xy_or_yx: str,
        show: Optional[bool] = True,
        aux_data: Optional[Union[None, Any]] = None,
        ttl_str: Optional[str] = "",
        x_axis_fontsize: Optional[float] = 25,
        y_axis_fontsize: Optional[float] = 25,
        ttl_fontsize: Optional[float] = 16,
        markersize: Optional[float] = 10,
        rho_ylims: Optional[list] = [10, 1000],
        phi_ylims: Optional[list] = [0, 90],
        figures_path: Optional[Union[str, pathlib.Path]] = FIGURES_PATH,
        figure_basename: Optional[str] = "",
    ) -> None:
        """
        One-off plotting method intended only for the synthetic test data for aurora dev

        Parameters
        ----------
        xy_or_yx: str
            Must be "xy" or "yx" -- the mode of the TF to plot
        show: Optional[bool]
            If True show figure on screen
        aux_data: Optional[Union[ZFile, None]]
            If not none, plot the data from the z-file as well as data from self
        ttl_str: Optional[str]
            If provided, ttl_str will be appended to the plot title
        x_axis_fontsize: int
            Font size for the x-axis
        y_axis_fontsize: int
            Font size for the y-axis
        ttl_fontsize: int
            Font size for the title
        markersize: int
            the marker size for the plot symbols
        rho_ylims: list
            The lower and upper limits for the y-axis on rho plot
        phi_ylims: list
            The lower and upper limits for the y-axis on phi plot
        figures_path: Union[str, pathlib.Path]
            Where to save the figure
        figure_basename:
            What the name of the figure will be.
            If not given, a default name is assigned

        """

        import matplotlib.pyplot as plt

        fig, axs = plt.subplots(nrows=2, figsize=(11, 8.5), dpi=300)

        color_cyc = {0: "r", 1: "orange", 2: "g", 3: "b", 4: "k"}
        for i_dec in self.tf_dict.keys():
            tf = self.tf_dict[i_dec]

            if xy_or_yx == "xy":
                aurora_rho = tf.rho[:, 0]
            else:
                aurora_rho = tf.rho[:, 1]
            plot_rho(
                axs[0],
                tf.periods,
                aurora_rho,
                marker="o",
                color=color_cyc[i_dec],
                linestyle="None",
                label=f"aurora {i_dec}",
                markersize=markersize,
            )

            if xy_or_yx == "xy":
                aurora_phi = tf.phi[:, 0]
            else:
                aurora_phi = tf.phi[:, 1]
            # rotate phases so all are positive:
            negative_phi_indices = np.where(aurora_phi < 0)[0]
            aurora_phi[negative_phi_indices] += 180.0
            plot_phi(
                axs[1],
                tf.periods,
                aurora_phi,
                marker="o",
                color=color_cyc[i_dec],
                linestyle="None",
                label=f"aurora {i_dec}",
                markersize=markersize,
            )

        if aux_data:
            decimation_levels = list(set(aux_data.decimation_levels))
            shape_cyc = {0: "s", 1: "v", 2: "*", 3: "^"}
            if xy_or_yx == "xy":
                emtf_rho = aux_data.rxy
                emtf_phi = aux_data.pxy
            elif xy_or_yx == "yx":
                emtf_rho = aux_data.ryx
                emtf_phi = aux_data.pyx

            axs[0].loglog(axs[0].get_xlim(), 100 * np.ones(2), color="k")
            axs[1].semilogx(axs[1].get_xlim(), 45 * np.ones(2), color="k")
            for i_dec in decimation_levels:
                ndx = np.where(aux_data.decimation_levels == i_dec)[0]
                axs[0].loglog(
                    aux_data.periods[ndx],
                    emtf_rho[ndx],
                    marker=shape_cyc[i_dec - 1],
                    color="k",
                    linestyle="None",
                    label=f"emtf " f"{int(i_dec-1)}",
                    markersize=markersize,
                )
                axs[1].semilogx(
                    aux_data.periods[ndx],
                    emtf_phi[ndx],
                    marker=shape_cyc[i_dec - 1],
                    color="k",
                    linestyle="None",
                    markersize=markersize,
                    label=f"emtf " f"{int(i_dec-1)}",
                )

        axs[0].legend(ncol=2)
        axs[1].legend(ncol=2)

        axs[1].set_xlabel("Period (s)", fontsize=x_axis_fontsize)
        axs[0].set_ylabel(r"$\Omega$-m", fontsize=y_axis_fontsize)
        axs[1].set_ylabel("Degrees", fontsize=y_axis_fontsize)

        ttl_str = f"{tf.tf_header.local_station.id} {xy_or_yx} \n{ttl_str}"
        axs[0].set_title(ttl_str, fontsize=ttl_fontsize)
        if rho_ylims is not None:
            axs[0].set_ylim(rho_ylims)
        if phi_ylims is not None:
            axs[1].set_ylim(phi_ylims)

        if figures_path is None:
            logger.info("figures path is not defined -- skipping saving figures")
            return
        else:
            if not figure_basename:
                figure_basename = f"{self.local_station_id}_{xy_or_yx}.png"
            out_file = figures_path.joinpath(figure_basename)
            plt.savefig(out_file)
            if show:
                plt.show()
            plt.close(fig)



"""
Tests for slicing RunTS objects using timestamps.

This will get moved into MTH5.
"""

import datetime

from loguru import logger
from mth5.utils.helpers import initialize_mth5


def test_can_slice_a_run_ts_using_timestamp(worker_safe_test1_h5, subtests):
    """Test that RunTS can be properly sliced using timestamps."""
    # Open the MTH5 file
    mth5_obj = initialize_mth5(worker_safe_test1_h5, "r")

    try:
        df = mth5_obj.channel_summary.to_dataframe()

        # Get the run
        try:
            run_001 = mth5_obj.get_run(station_name="test1", run_name="001")
        except ValueError:
            # This can happen on local machine
            run_001 = mth5_obj.get_run(
                station_name="test1",
                run_name="001",
                survey=mth5_obj.surveys_group.groups_list[0],
            )

        # Get the full run without slicing
        run_ts_full = run_001.to_runts()
        full_length = len(run_ts_full.dataset.ex.data)

        start = df.iloc[0].start
        end = df.iloc[0].end

        logger.info(f"Full run has {full_length} samples")
        logger.info(f"Start: {start}, End: {end}")

        # Test 1: Slice with exact start and end times
        with subtests.test(msg="exact_start_end"):
            run_ts_exact = run_001.to_runts(start=start, end=end)
            exact_length = len(run_ts_exact.dataset.ex.data)
            logger.info(f"Exact slice has {exact_length} samples")

            # Should have the same length as full run since we use exact bounds
            assert (
                exact_length == full_length
            ), f"Expected {full_length} samples with exact bounds, got {exact_length}"

        # Test 2: Slice with end + 499999 microseconds (less than one sample at 1 Hz)
        with subtests.test(msg="end_plus_499999_microseconds"):
            run_ts_sub_sample = run_001.to_runts(
                start=start, end=end + datetime.timedelta(microseconds=499999)
            )
            sub_sample_length = len(run_ts_sub_sample.dataset.ex.data)
            logger.info(f"End + 499999μs slice has {sub_sample_length} samples")

            # Should still have same length since we haven't crossed a sample boundary
            assert (
                sub_sample_length == full_length
            ), f"Expected {full_length} samples (sub-sample extension), got {sub_sample_length}"

        # Test 3: Slice with end + 500000 microseconds (half a sample at 1 Hz)
        with subtests.test(msg="end_plus_500000_microseconds"):
            run_ts_one_more = run_001.to_runts(
                start=start, end=end + datetime.timedelta(microseconds=500000)
            )
            one_more_length = len(run_ts_one_more.dataset.ex.data)
            logger.info(f"End + 500000μs slice has {one_more_length} samples")

            # The slicing appears to be inclusive of the exact end boundary
            # so adding 0.5 seconds doesn't add a new sample
            assert (
                one_more_length == full_length
            ), f"Expected {full_length} samples, got {one_more_length}"

        # Test 4: Verify that sliced data starts at correct time
        with subtests.test(msg="sliced_start_time"):
            run_ts_sliced = run_001.to_runts(start=start, end=end)
            sliced_start = run_ts_sliced.dataset.time.data[0]

            # Convert to comparable format - normalize timezones
            import pandas as pd

            expected_start = pd.Timestamp(start).tz_localize(None)
            actual_start = pd.Timestamp(sliced_start).tz_localize(None)

            logger.info(
                f"Expected start: {expected_start}, Actual start: {actual_start}"
            )
            assert (
                actual_start == expected_start
            ), f"Start time mismatch: expected {expected_start}, got {actual_start}"
    finally:
        mth5_obj.close_mth5()


def test_partial_run_slice(worker_safe_test1_h5):
    """Test slicing a partial section of a run."""
    # Open the MTH5 file
    mth5_obj = initialize_mth5(worker_safe_test1_h5, "r")

    try:
        df = mth5_obj.channel_summary.to_dataframe()

        # Get the run
        try:
            run_001 = mth5_obj.get_run(station_name="test1", run_name="001")
        except ValueError:
            run_001 = mth5_obj.get_run(
                station_name="test1",
                run_name="001",
                survey=mth5_obj.surveys_group.groups_list[0],
            )

        start = df.iloc[0].start
        end = df.iloc[0].end

        # Get full run
        run_ts_full = run_001.to_runts()
        full_length = len(run_ts_full.dataset.ex.data)

        # Slice the middle 50% of the run
        duration = end - start
        middle_start = start + duration * 0.25
        middle_end = start + duration * 0.75

        run_ts_middle = run_001.to_runts(start=middle_start, end=middle_end)
        middle_length = len(run_ts_middle.dataset.ex.data)

        logger.info(f"Full run: {full_length} samples")
        logger.info(f"Middle 50% slice: {middle_length} samples")

        # Middle section should be approximately 50% of full length
        # Allow for some tolerance due to rounding
        expected_middle = full_length * 0.5
        tolerance = full_length * 0.01  # 1% tolerance

        assert (
            abs(middle_length - expected_middle) <= tolerance
        ), f"Expected ~{expected_middle} samples in middle 50%, got {middle_length}"

        # Verify start time of sliced data
        import pandas as pd

        sliced_start = pd.Timestamp(run_ts_middle.dataset.time.data[0]).tz_localize(
            None
        )
        expected_start = pd.Timestamp(middle_start).tz_localize(None)

        assert (
            sliced_start == expected_start
        ), f"Start time mismatch: expected {expected_start}, got {sliced_start}"
    finally:
        mth5_obj.close_mth5()



"""
This module contains some helper functions that are called during the
execution of aurora's tests of processing on synthetic data.
"""

import pathlib
from typing import Optional, Union

import mt_metadata.transfer_functions
from mth5.data.make_mth5_from_asc import (
    create_test1_h5,
    create_test2_h5,
    create_test12rr_h5,
)

from aurora.pipelines.process_mth5 import process_mth5
from aurora.test_utils.synthetic.make_processing_configs import (
    make_processing_config_and_kernel_dataset,
)


def get_example_kernel_dataset(num_stations: int = 1):
    """
    Creates a kernel dataset object from the synthetic data
     - Helper function for synthetic tests.

    Returns
    -------
    kernel_dataset: aurora.transfer_function.kernel_dataset.KernelDataset
        The kernel dataset from a synthetic, single station mth5
    """

    from mth5.processing import KernelDataset, RunSummary

    if num_stations == 1:
        mth5_path = create_test1_h5(force_make_mth5=False)
    elif num_stations == 2:
        mth5_path = create_test12rr_h5(force_make_mth5=False)

    run_summary = RunSummary()
    run_summary.from_mth5s(
        [
            mth5_path,
        ]
    )

    kernel_dataset = KernelDataset()
    station_id = run_summary.df.station.iloc[0]
    if num_stations == 1:
        kernel_dataset.from_run_summary(
            run_summary=run_summary, local_station_id=station_id
        )
    elif num_stations == 2:
        remote_station_id = run_summary.df.station.iloc[1]
        kernel_dataset.from_run_summary(
            run_summary=run_summary,
            local_station_id=station_id,
            remote_station_id=remote_station_id,
        )
    return kernel_dataset


def tf_obj_from_synthetic_data(
    mth5_path: pathlib.Path,
) -> mt_metadata.transfer_functions.TF:
    """
    Executes aurora processing on mth5_path, and returns mt_metadata TF object.
    - Helper function for test_issue_139

    """
    from mth5.processing import KernelDataset, RunSummary

    from aurora.config.config_creator import ConfigCreator

    run_summary = RunSummary()
    run_summary.from_mth5s(list((mth5_path,)))

    kernel_dataset = KernelDataset()
    kernel_dataset.from_run_summary(run_summary, "test1", "test2")

    # Define the processing Configuration
    cc = ConfigCreator()
    config = cc.create_from_kernel_dataset(kernel_dataset)

    tf_cls = process_mth5(
        config,
        kernel_dataset,
        units="MT",
        z_file_path="test1_RRtest2.zrr",
    )
    return tf_cls


def process_synthetic_1(
    config_keyword: Optional[str] = "test1",
    z_file_path: Optional[Union[str, pathlib.Path]] = "",
    test_scale_factor: Optional[bool] = False,
    simultaneous_regression: Optional[bool] = False,
    file_version: Optional[str] = "0.1.0",  # TODO: set to Literal["0.1.0", "0.2.0"]
    return_collection: Optional[bool] = False,
    channel_nomenclature: Optional[str] = "default",
    reload_config: Optional[bool] = False,
    mth5_path: Optional[Union[str, pathlib.Path]] = None,
):
    """

    Parameters
    ----------
    config_keyword: str
        "test1", "test1_tfk", this is an argument passed to the create_test_run_config
        as test_case_id.
    z_file_path: str or path
        Where the z-file will be output
    test_scale_factor: bool
        If true, will assign scale factors to the channels
    simultaneous_regression: bool
        If True will do regression all outut channels in one step, rather than the
        usual, channel-by-channel method
    file_version: str
        one of ["0.1.0", "0.2.0"]
    mth5_path: str or path, optional
        Path to an existing test1.h5 MTH5 file. If None, will create one.

    Returns
    -------
    tf_result: TransferFunctionCollection or mt_metadata.transfer_functions.TF
        Should change so that it is mt_metadata.TF (see Issue #143)
    """
    if mth5_path is None:
        mth5_path = create_test1_h5(
            file_version=file_version, channel_nomenclature=channel_nomenclature
        )
    mth5_paths = [
        mth5_path,
    ]
    station_id = "test1"
    tfk_dataset, processing_config = make_processing_config_and_kernel_dataset(
        config_keyword=station_id,
        station_id=station_id,
        remote_id=None,  # TODO: allow empty str instead of None
        mth5s=mth5_paths,
        channel_nomenclature=channel_nomenclature,
    )

    # Test that channel_scale_factors column is optional
    if test_scale_factor:
        scale_factors = {
            "ex": 10.0,
            "ey": 3.0,
            "hx": 6.0,
            "hy": 5.0,
            "hz": 100.0,
        }
        tfk_dataset.df.at[0, "channel_scale_factors"] = scale_factors
    else:
        tfk_dataset.df.drop(columns=["channel_scale_factors"], inplace=True)

    # Relates to issue #172
    # reload_config = True
    # if reload_config:
    #     from mt_metadata.processing.aurora import Processing
    #     p = Processing()
    #     config_path = pathlib.Path("config")
    #     json_fn = config_path.joinpath(processing_config.json_fn())
    #     p.from_json(json_fn)

    if simultaneous_regression:
        for decimation in processing_config.decimations:
            decimation.estimator.estimate_per_channel = False

    tf_result = process_mth5(
        processing_config,
        tfk_dataset=tfk_dataset,
        z_file_path=z_file_path,
        return_collection=return_collection,
    )

    if return_collection:
        z_figure_name = z_file_path.name.replace("zss", "png")
        for xy_or_yx in ["xy", "yx"]:
            ttl_str = f"{xy_or_yx} component, test_scale_factor = {test_scale_factor}"
            out_png_name = f"{xy_or_yx}_{z_figure_name}"
            tf_result.rho_phi_plot(
                xy_or_yx=xy_or_yx,
                ttl_str=ttl_str,
                show=False,
                figure_basename=out_png_name,
                figures_path=z_file_path.parent,  # TODO: check this works
            )
    return tf_result


def process_synthetic_2(
    force_make_mth5: Optional[bool] = True,
    z_file_path: Optional[Union[str, pathlib.Path, None]] = None,
    save_fc: Optional[bool] = False,
    file_version: Optional[str] = "0.2.0",
    channel_nomenclature: Optional[str] = "default",
    mth5_path: Optional[Union[str, pathlib.Path]] = None,
):
    """"""
    station_id = "test2"
    if mth5_path is None:
        mth5_path = create_test2_h5(
            force_make_mth5=force_make_mth5, file_version=file_version
        )
    mth5_paths = [
        mth5_path,
    ]

    tfk_dataset, processing_config = make_processing_config_and_kernel_dataset(
        config_keyword=station_id,
        station_id=station_id,
        remote_id=None,
        mth5s=mth5_paths,
        channel_nomenclature=channel_nomenclature,
    )

    for decimation_level in processing_config.decimations:
        if save_fc:
            decimation_level.save_fcs = True
            decimation_level.save_fcs_type = "h5"  # h5 instead of "csv"

    tfc = process_mth5(
        processing_config,
        tfk_dataset=tfk_dataset,
        z_file_path=z_file_path,
    )
    return tfc


def process_synthetic_1r2(
    config_keyword="test1r2",
    channel_nomenclature="default",
    return_collection=False,
    mth5_path: Optional[Union[str, pathlib.Path]] = None,
):
    if mth5_path is None:
        mth5_path = create_test12rr_h5(channel_nomenclature=channel_nomenclature)
    mth5_paths = [
        mth5_path,
    ]

    tfk_dataset, processing_config = make_processing_config_and_kernel_dataset(
        config_keyword,
        station_id="test1",
        remote_id="test2",
        mth5s=mth5_paths,
        channel_nomenclature=channel_nomenclature,
    )

    tfc = process_mth5(
        processing_config,
        tfk_dataset=tfk_dataset,
        return_collection=return_collection,
    )
    return tfc



# -*- coding: utf-8 -*-
"""
Pytest suite for cross_power module.

Tests transfer function computation from covariance matrices using fixtures
and subtests where appropriate. Optimized for pytest-xdist parallel execution.
"""

import numpy as np
import pytest
from mt_metadata.transfer_functions import (
    STANDARD_INPUT_CHANNELS,
    STANDARD_OUTPUT_CHANNELS,
)
from mth5.timeseries.xarray_helpers import initialize_xrda_2d_cov

from aurora.transfer_function.cross_power import (
    _channel_names,
    _tf__x,
    _tf__y,
    _tx,
    _ty,
    _zxx,
    _zxy,
    _zyx,
    _zyy,
    tf_from_cross_powers,
)


# =============================================================================
# Fixtures
# =============================================================================


@pytest.fixture(scope="module")
def station_ids():
    """Station IDs for testing."""
    return ["MT1", "MT2"]


@pytest.fixture(scope="module")
def components():
    """Standard MT components."""
    return STANDARD_INPUT_CHANNELS + STANDARD_OUTPUT_CHANNELS


@pytest.fixture(scope="module")
def channel_labels(station_ids, components):
    """Generate channel labels for both stations."""
    station_1_channels = [f"{station_ids[0]}_{x}" for x in components]
    station_2_channels = [f"{station_ids[1]}_{x}" for x in components]
    return station_1_channels + station_2_channels


@pytest.fixture(scope="module")
def sdm_covariance(channel_labels):
    """
    Create a synthetic covariance matrix for testing.

    Uses module scope for efficiency with pytest-xdist.
    """
    sdm = initialize_xrda_2d_cov(
        channels=channel_labels,
        dtype=complex,
    )
    np.random.seed(0)
    data = np.random.random((len(channel_labels), 1000))
    sdm.data = np.cov(data)
    return sdm


@pytest.fixture(scope="module")
def simple_sdm():
    """
    Create a simple 2x2 covariance matrix for unit testing.

    This allows testing specific mathematical properties without
    the complexity of the full covariance matrix.
    """
    channels = ["MT1_hx", "MT1_hy"]
    sdm = initialize_xrda_2d_cov(channels=channels, dtype=complex)
    # Create a simple hermitian matrix
    sdm.data = np.array([[2.0 + 0j, 1.0 + 0.5j], [1.0 - 0.5j, 3.0 + 0j]])
    return sdm


@pytest.fixture(scope="module")
def identity_sdm():
    """Create an identity-like covariance matrix for edge case testing."""
    channels = ["MT1_ex", "MT1_ey", "MT1_hx", "MT1_hy", "MT1_hz"]
    sdm = initialize_xrda_2d_cov(channels=channels, dtype=complex)
    sdm.data = np.eye(len(channels), dtype=complex)
    return sdm


@pytest.fixture
def channel_names_fixture(station_ids):
    """Fixture providing channel names for a single station."""
    station = station_ids[0]
    remote = station_ids[1]
    return _channel_names(station_id=station, remote=remote, join_char="_")


# =============================================================================
# Test Channel Names
# =============================================================================


class TestChannelNames:
    """Test channel name generation with different configurations."""

    def test_channel_names_with_remote(self, station_ids):
        """Test channel name generation with remote reference."""
        station = station_ids[0]
        remote = station_ids[1]
        Ex, Ey, Hx, Hy, Hz, A, B = _channel_names(
            station_id=station, remote=remote, join_char="_"
        )
        assert Ex == f"{station}_ex"
        assert Ey == f"{station}_ey"
        assert Hx == f"{station}_hx"
        assert Hy == f"{station}_hy"
        assert Hz == f"{station}_hz"
        assert A == f"{remote}_hx"
        assert B == f"{remote}_hy"

    def test_channel_names_without_remote(self, station_ids):
        """Test channel name generation for single station (no remote)."""
        station = station_ids[0]
        Ex, Ey, Hx, Hy, Hz, A, B = _channel_names(
            station_id=station, remote="", join_char="_"
        )
        assert Ex == f"{station}_ex"
        assert Ey == f"{station}_ey"
        assert Hx == f"{station}_hx"
        assert Hy == f"{station}_hy"
        assert Hz == f"{station}_hz"
        # For single station, A and B should use station's own channels
        assert A == f"{station}_hx"
        assert B == f"{station}_hy"

    def test_channel_names_custom_join_char(self, station_ids):
        """Test channel names with custom join character."""
        station = station_ids[0]
        remote = station_ids[1]
        Ex, Ey, Hx, Hy, Hz, A, B = _channel_names(
            station_id=station, remote=remote, join_char="-"
        )
        assert Ex == f"{station}-ex"
        assert Ey == f"{station}-ey"
        assert Hx == f"{station}-hx"
        assert Hy == f"{station}-hy"
        assert Hz == f"{station}-hz"
        assert A == f"{remote}-hx"
        assert B == f"{remote}-hy"

    def test_channel_names_return_type(self, station_ids):
        """Test that _channel_names returns a tuple of 7 elements."""
        result = _channel_names(
            station_id=station_ids[0], remote=station_ids[1], join_char="_"
        )
        assert isinstance(result, tuple)
        assert len(result) == 7
        assert all(isinstance(name, str) for name in result)


# =============================================================================
# Test Transfer Function Computation
# =============================================================================


class TestTFComputationBasic:
    """Test basic transfer function element computations."""

    def test_tf__x_computation(self, sdm_covariance, channel_names_fixture):
        """Test _tf__x function computes without error."""
        Ex, Ey, Hx, Hy, Hz, A, B = channel_names_fixture
        result = _tf__x(sdm_covariance, Y=Ex, Hx=Hx, Hy=Hy, A=A, B=B)
        # Result may be xarray DataArray, extract value
        value = result.item() if hasattr(result, "item") else result
        assert isinstance(value, (complex, np.complexfloating, float, np.floating))

    def test_tf__y_computation(self, sdm_covariance, channel_names_fixture):
        """Test _tf__y function computes without error."""
        Ex, Ey, Hx, Hy, Hz, A, B = channel_names_fixture
        result = _tf__y(sdm_covariance, Y=Ex, Hx=Hx, Hy=Hy, A=A, B=B)
        # Result may be xarray DataArray, extract value
        value = result.item() if hasattr(result, "item") else result
        assert isinstance(value, (complex, np.complexfloating, float, np.floating))

    def test_zxx_computation(self, sdm_covariance, channel_names_fixture):
        """Test _zxx function computes without error."""
        Ex, Ey, Hx, Hy, Hz, A, B = channel_names_fixture
        result = _zxx(sdm_covariance, Ex=Ex, Hx=Hx, Hy=Hy, A=A, B=B)
        # Result may be xarray DataArray, extract value
        value = result.item() if hasattr(result, "item") else result
        assert isinstance(value, (complex, np.complexfloating, float, np.floating))

    def test_zxy_computation(self, sdm_covariance, channel_names_fixture):
        """Test _zxy function computes without error."""
        Ex, Ey, Hx, Hy, Hz, A, B = channel_names_fixture
        result = _zxy(sdm_covariance, Ex=Ex, Hx=Hx, Hy=Hy, A=A, B=B)
        # Result may be xarray DataArray, extract value
        value = result.item() if hasattr(result, "item") else result
        assert isinstance(value, (complex, np.complexfloating, float, np.floating))

    def test_zyx_computation(self, sdm_covariance, channel_names_fixture):
        """Test _zyx function computes without error."""
        Ex, Ey, Hx, Hy, Hz, A, B = channel_names_fixture
        result = _zyx(sdm_covariance, Ey=Ey, Hx=Hx, Hy=Hy, A=A, B=B)
        # Result may be xarray DataArray, extract value
        value = result.item() if hasattr(result, "item") else result
        assert isinstance(value, (complex, np.complexfloating, float, np.floating))

    def test_zyy_computation(self, sdm_covariance, channel_names_fixture):
        """Test _zyy function computes without error."""
        Ex, Ey, Hx, Hy, Hz, A, B = channel_names_fixture
        result = _zyy(sdm_covariance, Ey=Ey, Hx=Hx, Hy=Hy, A=A, B=B)
        # Result may be xarray DataArray, extract value
        value = result.item() if hasattr(result, "item") else result
        assert isinstance(value, (complex, np.complexfloating, float, np.floating))

    def test_tx_computation(self, sdm_covariance, channel_names_fixture):
        """Test _tx function computes without error."""
        Ex, Ey, Hx, Hy, Hz, A, B = channel_names_fixture
        result = _tx(sdm_covariance, Hz=Hz, Hx=Hx, Hy=Hy, A=A, B=B)
        # Result may be xarray DataArray, extract value
        value = result.item() if hasattr(result, "item") else result
        assert isinstance(value, (complex, np.complexfloating, float, np.floating))

    def test_ty_computation(self, sdm_covariance, channel_names_fixture):
        """Test _ty function computes without error."""
        Ex, Ey, Hx, Hy, Hz, A, B = channel_names_fixture
        result = _ty(sdm_covariance, Hz=Hz, Hx=Hx, Hy=Hy, A=A, B=B)
        # Result may be xarray DataArray, extract value
        value = result.item() if hasattr(result, "item") else result
        assert isinstance(value, (complex, np.complexfloating, float, np.floating))


class TestVozoffEquations:
    """Test Vozoff equation equivalences and generalizations."""

    def test_generalizing_vozoffs_equations(
        self, sdm_covariance, channel_names_fixture
    ):
        """
        Test that specific Vozoff equations match generalized formulations.

        Verifies that _zxx, _zxy, _zyx, _zyy, _tx, _ty are equivalent to
        _tf__x and _tf__y with appropriate parameters.
        """
        Ex, Ey, Hx, Hy, Hz, A, B = channel_names_fixture

        # Test impedance tensor elements
        assert _zxx(sdm_covariance, Ex=Ex, Hx=Hx, Hy=Hy, A=A, B=B) == _tf__x(
            sdm_covariance, Y=Ex, Hx=Hx, Hy=Hy, A=A, B=B
        )
        assert _zxy(sdm_covariance, Ex=Ex, Hx=Hx, Hy=Hy, A=A, B=B) == _tf__y(
            sdm_covariance, Y=Ex, Hx=Hx, Hy=Hy, A=A, B=B
        )
        assert _zyx(sdm_covariance, Ey=Ey, Hx=Hx, Hy=Hy, A=A, B=B) == _tf__x(
            sdm_covariance, Y=Ey, Hx=Hx, Hy=Hy, A=A, B=B
        )
        assert _zyy(sdm_covariance, Ey=Ey, Hx=Hx, Hy=Hy, A=A, B=B) == _tf__y(
            sdm_covariance, Y=Ey, Hx=Hx, Hy=Hy, A=A, B=B
        )

        # Test tipper elements
        assert _tx(sdm_covariance, Hz=Hz, Hx=Hx, Hy=Hy, A=A, B=B) == _tf__x(
            sdm_covariance, Y=Hz, Hx=Hx, Hy=Hy, A=A, B=B
        )
        assert _ty(sdm_covariance, Hz=Hz, Hx=Hx, Hy=Hy, A=A, B=B) == _tf__y(
            sdm_covariance, Y=Hz, Hx=Hx, Hy=Hy, A=A, B=B
        )

    def test_impedance_symmetry(self, sdm_covariance, channel_names_fixture):
        """
        Test symmetry properties of impedance tensor.

        Verifies that Ex->Ey substitution relates Z_xx to Z_yx and Z_xy to Z_yy.
        """
        Ex, Ey, Hx, Hy, Hz, A, B = channel_names_fixture

        # Z_xx with Ex should have same structure as Z_yx with Ey
        zxx_result = _tf__x(sdm_covariance, Y=Ex, Hx=Hx, Hy=Hy, A=A, B=B)
        zyx_result = _tf__x(sdm_covariance, Y=Ey, Hx=Hx, Hy=Hy, A=A, B=B)

        # Both should be numeric (extract values if DataArray)
        zxx_val = zxx_result.item() if hasattr(zxx_result, "item") else zxx_result
        zyx_val = zyx_result.item() if hasattr(zyx_result, "item") else zyx_result
        assert isinstance(zxx_val, (complex, np.complexfloating, float, np.floating))
        assert isinstance(zyx_val, (complex, np.complexfloating, float, np.floating))

        # Z_xy with Ex should have same structure as Z_yy with Ey
        zxy_result = _tf__y(sdm_covariance, Y=Ex, Hx=Hx, Hy=Hy, A=A, B=B)
        zyy_result = _tf__y(sdm_covariance, Y=Ey, Hx=Hx, Hy=Hy, A=A, B=B)

        zxy_val = zxy_result.item() if hasattr(zxy_result, "item") else zxy_result
        zyy_val = zyy_result.item() if hasattr(zyy_result, "item") else zyy_result
        assert isinstance(zxy_val, (complex, np.complexfloating, float, np.floating))
        assert isinstance(zyy_val, (complex, np.complexfloating, float, np.floating))


class TestTFFromCrossPowers:
    """Test the main tf_from_cross_powers function."""

    def test_tf_from_cross_powers_dict_output(self, sdm_covariance, station_ids):
        """Test tf_from_cross_powers returns dictionary with all components."""
        result = tf_from_cross_powers(
            sdm_covariance,
            station_id=station_ids[0],
            remote=station_ids[1],
            output_format="dict",
        )

        assert isinstance(result, dict)
        expected_keys = ["z_xx", "z_xy", "z_yx", "z_yy", "t_zx", "t_zy"]
        assert set(result.keys()) == set(expected_keys)

        # All values should be numeric (may be wrapped in DataArray)
        for key, value in result.items():
            val = value.item() if hasattr(value, "item") else value
            assert isinstance(val, (complex, np.complexfloating, float, np.floating))

    def test_tf_from_cross_powers_single_station(self, sdm_covariance, station_ids):
        """Test tf_from_cross_powers without remote reference."""
        result = tf_from_cross_powers(
            sdm_covariance,
            station_id=station_ids[0],
            remote="",
            output_format="dict",
        )

        assert isinstance(result, dict)
        expected_keys = ["z_xx", "z_xy", "z_yx", "z_yy", "t_zx", "t_zy"]
        assert set(result.keys()) == set(expected_keys)

    def test_tf_from_cross_powers_mt_metadata_format(self, sdm_covariance, station_ids):
        """Test that mt_metadata format raises NotImplementedError."""
        with pytest.raises(NotImplementedError):
            tf_from_cross_powers(
                sdm_covariance,
                station_id=station_ids[0],
                remote=station_ids[1],
                output_format="mt_metadata",
            )


# =============================================================================
# Test Mathematical Properties
# =============================================================================


class TestMathematicalProperties:
    """Test mathematical properties of transfer function computations."""

    def test_hermitian_symmetry(self, sdm_covariance, channel_names_fixture):
        """
        Test that covariance matrix hermitian symmetry is respected.

        For a hermitian matrix, sdm[i,j] = conj(sdm[j,i])
        """
        Ex, Ey, Hx, Hy, Hz, A, B = channel_names_fixture

        # Check a few elements for hermitian symmetry
        assert np.isclose(
            sdm_covariance.loc[Ex, Hx], np.conj(sdm_covariance.loc[Hx, Ex])
        )
        assert np.isclose(
            sdm_covariance.loc[Ey, Hy], np.conj(sdm_covariance.loc[Hy, Ey])
        )

    def test_denominator_consistency(self, sdm_covariance, channel_names_fixture):
        """
        Test that denominators are consistent across related TF elements.

        Z_xx and Z_yx share the same denominator: <Hx,A*><Hy,B*> - <Hx,B*><Hy,A*>
        Z_xy and Z_yy share the same denominator: <Hy,A*><Hx,B*> - <Hy,B*><Hx,A*>
        """
        Ex, Ey, Hx, Hy, Hz, A, B = channel_names_fixture

        # Compute shared denominator for Z_xx and Z_yx
        denom_x = (
            sdm_covariance.loc[Hx, A] * sdm_covariance.loc[Hy, B]
            - sdm_covariance.loc[Hx, B] * sdm_covariance.loc[Hy, A]
        )

        # Compute shared denominator for Z_xy and Z_yy
        denom_y = (
            sdm_covariance.loc[Hy, A] * sdm_covariance.loc[Hx, B]
            - sdm_covariance.loc[Hy, B] * sdm_covariance.loc[Hx, A]
        )

        # Both denominators should be non-zero for well-conditioned matrices
        assert not np.isclose(denom_x, 0)
        assert not np.isclose(denom_y, 0)

    def test_tf_finite_values(self, sdm_covariance, channel_names_fixture):
        """Test that computed TF values are finite (not NaN or inf)."""
        Ex, Ey, Hx, Hy, Hz, A, B = channel_names_fixture

        # Test all TF components
        tf_values = [
            _zxx(sdm_covariance, Ex=Ex, Hx=Hx, Hy=Hy, A=A, B=B),
            _zxy(sdm_covariance, Ex=Ex, Hx=Hx, Hy=Hy, A=A, B=B),
            _zyx(sdm_covariance, Ey=Ey, Hx=Hx, Hy=Hy, A=A, B=B),
            _zyy(sdm_covariance, Ey=Ey, Hx=Hx, Hy=Hy, A=A, B=B),
            _tx(sdm_covariance, Hz=Hz, Hx=Hx, Hy=Hy, A=A, B=B),
            _ty(sdm_covariance, Hz=Hz, Hx=Hx, Hy=Hy, A=A, B=B),
        ]

        for value in tf_values:
            assert np.isfinite(value)


# =============================================================================
# Test Edge Cases
# =============================================================================


class TestEdgeCases:
    """Test edge cases and boundary conditions."""

    def test_identity_covariance_matrix(self, identity_sdm):
        """Test TF computation with identity-like covariance matrix."""
        station = "MT1"
        Ex, Ey, Hx, Hy, Hz, A, B = _channel_names(
            station_id=station, remote="", join_char="_"
        )

        # With identity matrix, many cross terms are zero
        # Denominator: <Hx,Hx><Hy,Hy> - <Hx,Hy><Hy,Hx> = 1*1 - 0*0 = 1
        denom_x = (
            identity_sdm.loc[Hx, A] * identity_sdm.loc[Hy, B]
            - identity_sdm.loc[Hx, B] * identity_sdm.loc[Hy, A]
        )
        assert np.isclose(denom_x, 1.0)

    def test_different_join_characters(self, sdm_covariance, station_ids, subtests):
        """Test TF computation with different join characters."""
        join_chars = ["_", "-", ".", ""]

        for join_char in join_chars:
            with subtests.test(join_char=join_char):
                # This will fail for non-underscore join chars since our
                # sdm_covariance fixture uses underscore
                # But test the function interface
                Ex, Ey, Hx, Hy, Hz, A, B = _channel_names(
                    station_id=station_ids[0],
                    remote=station_ids[1],
                    join_char=join_char,
                )

                # Verify the join character is used
                assert join_char in Ex or join_char == ""
                assert Ex.startswith(station_ids[0])

    def test_zero_cross_power_handling(self):
        """Test behavior when some cross-power terms are zero."""
        channels = ["MT1_ex", "MT1_hx", "MT1_hy", "MT2_hx", "MT2_hy"]
        sdm = initialize_xrda_2d_cov(channels=channels, dtype=complex)

        # Create a matrix where some cross terms are zero
        sdm.data = np.eye(len(channels), dtype=complex)
        # Add some non-zero diagonal elements
        sdm.data[0, 0] = 2.0
        sdm.data[1, 1] = 3.0
        sdm.data[2, 2] = 4.0

        Ex, Ey, Hx, Hy, Hz, A, B = _channel_names(
            station_id="MT1", remote="MT2", join_char="_"
        )

        # Should compute without error even with many zeros
        result = _tf__x(sdm, Y=Ex, Hx=Hx, Hy=Hy, A=A, B=B)
        val = result.item() if hasattr(result, "item") else result
        # Result might be NaN due to zero denominator, that's OK
        assert isinstance(val, (complex, np.complexfloating, float, np.floating))


# =============================================================================
# Test Data Integrity
# =============================================================================


class TestDataIntegrity:
    """Test that TF computation doesn't modify input data."""

    def test_input_sdm_unchanged(self, sdm_covariance, station_ids):
        """Test that tf_from_cross_powers doesn't modify input covariance matrix."""
        # Make a copy of the original data
        original_data = sdm_covariance.data.copy()

        # Compute TF
        tf_from_cross_powers(
            sdm_covariance,
            station_id=station_ids[0],
            remote=station_ids[1],
        )

        # Verify data unchanged
        assert np.allclose(sdm_covariance.data, original_data)

    def test_individual_tf_functions_unchanged(
        self, sdm_covariance, channel_names_fixture
    ):
        """Test that individual TF functions don't modify input."""
        Ex, Ey, Hx, Hy, Hz, A, B = channel_names_fixture
        original_data = sdm_covariance.data.copy()

        # Call all TF functions
        _zxx(sdm_covariance, Ex=Ex, Hx=Hx, Hy=Hy, A=A, B=B)
        _zxy(sdm_covariance, Ex=Ex, Hx=Hx, Hy=Hy, A=A, B=B)
        _zyx(sdm_covariance, Ey=Ey, Hx=Hx, Hy=Hy, A=A, B=B)
        _zyy(sdm_covariance, Ey=Ey, Hx=Hx, Hy=Hy, A=A, B=B)
        _tx(sdm_covariance, Hz=Hz, Hx=Hx, Hy=Hy, A=A, B=B)
        _ty(sdm_covariance, Hz=Hz, Hx=Hx, Hy=Hy, A=A, B=B)

        # Verify data unchanged
        assert np.allclose(sdm_covariance.data, original_data)


# =============================================================================
# Test Numerical Stability
# =============================================================================


class TestNumericalStability:
    """Test numerical stability with various input conditions."""

    def test_small_values_stability(self):
        """Test TF computation with very small covariance values."""
        channels = ["MT1_ex", "MT1_hx", "MT1_hy", "MT2_hx", "MT2_hy"]
        sdm = initialize_xrda_2d_cov(channels=channels, dtype=complex)

        # Create matrix with small values
        np.random.seed(42)
        sdm.data = np.random.random((len(channels), len(channels))) * 1e-10
        sdm.data = sdm.data + sdm.data.T.conj()  # Make hermitian

        Ex, Ey, Hx, Hy, Hz, A, B = _channel_names(
            station_id="MT1", remote="MT2", join_char="_"
        )

        result = _tf__x(sdm, Y=Ex, Hx=Hx, Hy=Hy, A=A, B=B)
        # Result might be large due to small denominator, but should be finite
        assert np.isfinite(result) or np.isinf(result)  # Allow inf for edge case

    def test_large_values_stability(self):
        """Test TF computation with very large covariance values."""
        channels = ["MT1_ex", "MT1_hx", "MT1_hy", "MT2_hx", "MT2_hy"]
        sdm = initialize_xrda_2d_cov(channels=channels, dtype=complex)

        # Create matrix with large values
        np.random.seed(43)
        sdm.data = np.random.random((len(channels), len(channels))) * 1e10
        sdm.data = sdm.data + sdm.data.T.conj()  # Make hermitian

        Ex, Ey, Hx, Hy, Hz, A, B = _channel_names(
            station_id="MT1", remote="MT2", join_char="_"
        )

        result = _tf__x(sdm, Y=Ex, Hx=Hx, Hy=Hy, A=A, B=B)
        assert np.isfinite(result)

    def test_complex_phase_variations(self, subtests):
        """Test TF computation with various complex phase relationships."""
        channels = ["MT1_ex", "MT1_hx", "MT1_hy", "MT2_hx", "MT2_hy"]

        phases = [0, np.pi / 4, np.pi / 2, np.pi, 3 * np.pi / 2]

        for phase in phases:
            with subtests.test(phase=phase):
                sdm = initialize_xrda_2d_cov(channels=channels, dtype=complex)

                # Create matrix with specific phase
                np.random.seed(44)
                magnitude = np.random.random((len(channels), len(channels)))
                sdm.data = magnitude * np.exp(1j * phase)
                sdm.data = sdm.data + sdm.data.T.conj()  # Make hermitian

                Ex, Ey, Hx, Hy, Hz, A, B = _channel_names(
                    station_id="MT1", remote="MT2", join_char="_"
                )

                result = _tf__x(sdm, Y=Ex, Hx=Hx, Hy=Hy, A=A, B=B)
                val = result.item() if hasattr(result, "item") else result
                assert isinstance(
                    val, (complex, np.complexfloating, float, np.floating)
                )


# =============================================================================
# Test Return Value Characteristics
# =============================================================================


class TestReturnValues:
    """Test characteristics of return values from TF functions."""

    def test_all_tf_components_present(self, sdm_covariance, station_ids):
        """Test that tf_from_cross_powers returns all expected components."""
        result = tf_from_cross_powers(
            sdm_covariance,
            station_id=station_ids[0],
            remote=station_ids[1],
        )

        # Check all standard TF components are present
        assert "z_xx" in result
        assert "z_xy" in result
        assert "z_yx" in result
        assert "z_yy" in result
        assert "t_zx" in result
        assert "t_zy" in result

        # Should only have these 6 components
        assert len(result) == 6

    def test_tf_component_types(self, sdm_covariance, station_ids):
        """Test that all TF components are complex numbers."""
        result = tf_from_cross_powers(
            sdm_covariance,
            station_id=station_ids[0],
            remote=station_ids[1],
        )

        for component_name, value in result.items():
            val = value.item() if hasattr(value, "item") else value
            assert isinstance(
                val, (complex, np.complexfloating, float, np.floating)
            ), f"{component_name} is not numeric"

    def test_impedance_vs_tipper_separation(self, sdm_covariance, station_ids):
        """Test that impedance and tipper components are computed separately."""
        result = tf_from_cross_powers(
            sdm_covariance,
            station_id=station_ids[0],
            remote=station_ids[1],
        )

        impedance_keys = ["z_xx", "z_xy", "z_yx", "z_yy"]
        tipper_keys = ["t_zx", "t_zy"]

        # All impedance components should be present
        for key in impedance_keys:
            assert key in result

        # All tipper components should be present
        for key in tipper_keys:
            assert key in result


# =============================================================================
# Test Consistency Across Calls
# =============================================================================


class TestConsistency:
    """Test consistency of results across multiple calls."""

    def test_deterministic_results(self, sdm_covariance, station_ids):
        """Test that repeated calls produce identical results."""
        result1 = tf_from_cross_powers(
            sdm_covariance,
            station_id=station_ids[0],
            remote=station_ids[1],
        )

        result2 = tf_from_cross_powers(
            sdm_covariance,
            station_id=station_ids[0],
            remote=station_ids[1],
        )

        for key in result1.keys():
            assert result1[key] == result2[key]

    def test_individual_function_consistency(
        self, sdm_covariance, channel_names_fixture
    ):
        """Test that individual TF functions produce consistent results."""
        Ex, Ey, Hx, Hy, Hz, A, B = channel_names_fixture

        # Call the same function multiple times
        results = [
            _zxx(sdm_covariance, Ex=Ex, Hx=Hx, Hy=Hy, A=A, B=B) for _ in range(5)
        ]

        # All results should be identical
        for result in results[1:]:
            assert result == results[0]



"""
Pytest suite for testing xarray_helpers module.

Tests cover:
- nan_to_mean: Replacing NaN values with channel means
- handle_nan: Dropping NaN values across multiple datasets
- time_axis_match: Checking time coordinate alignment
- Edge cases and parameter variations

Optimized for pytest-xdist parallel execution.
"""

import numpy as np
import pytest
import xarray as xr

from aurora.time_series.xarray_helpers import handle_nan, nan_to_mean, time_axis_match


# =============================================================================
# Fixtures
# =============================================================================


@pytest.fixture
def basic_times():
    """Basic time coordinate array."""
    return np.array([0, 1, 2, 3])


@pytest.fixture
def extended_times():
    """Extended time coordinate array for edge case testing."""
    return np.array([0, 1, 2, 3, 4])


@pytest.fixture
def single_channel_dataset_with_nan(basic_times):
    """Dataset with single channel containing NaN values."""
    data = np.array([1.0, np.nan, 3.0, 4.0])
    return xr.Dataset({"hx": ("time", data)}, coords={"time": basic_times})


@pytest.fixture
def multi_channel_dataset_with_nan(basic_times):
    """Dataset with multiple channels containing NaN values in different locations."""
    data_hx = np.array([1.0, np.nan, 3.0, 4.0])
    data_hy = np.array([np.nan, 2.0, 3.0, 4.0])
    return xr.Dataset(
        {
            "hx": ("time", data_hx),
            "hy": ("time", data_hy),
        },
        coords={"time": basic_times},
    )


@pytest.fixture
def dataset_no_nan(basic_times):
    """Dataset without any NaN values."""
    data = np.array([1.0, 2.0, 3.0, 4.0])
    return xr.Dataset({"hx": ("time", data)}, coords={"time": basic_times})


@pytest.fixture
def dataset_all_nan(basic_times):
    """Dataset with all NaN values."""
    data = np.array([np.nan, np.nan, np.nan, np.nan])
    return xr.Dataset({"hx": ("time", data)}, coords={"time": basic_times})


# =============================================================================
# Test Classes
# =============================================================================


class TestNanToMean:
    """Test nan_to_mean function."""

    def test_nan_to_mean_basic(self, single_channel_dataset_with_nan):
        """Test nan_to_mean replaces NaNs with mean per channel."""
        ds_filled = nan_to_mean(single_channel_dataset_with_nan.copy())

        # The mean ignoring NaN is (1+3+4)/3 = 2.666...
        expected = np.array([1.0, 2.66666667, 3.0, 4.0])
        assert np.allclose(ds_filled.hx.values, expected)

        # No NaNs should remain
        assert not np.any(np.isnan(ds_filled.hx.values))

    def test_nan_to_mean_multiple_channels(self, multi_channel_dataset_with_nan):
        """Test nan_to_mean with multiple channels and NaNs in different places."""
        ds_filled = nan_to_mean(multi_channel_dataset_with_nan.copy())

        expected_hx = np.array([1.0, 2.66666667, 3.0, 4.0])
        expected_hy = np.array([3.0, 2.0, 3.0, 4.0])

        assert np.allclose(ds_filled.hx.values, expected_hx)
        assert np.allclose(ds_filled.hy.values, expected_hy)
        assert not np.any(np.isnan(ds_filled.hx.values))
        assert not np.any(np.isnan(ds_filled.hy.values))

    def test_nan_to_mean_no_nans(self, dataset_no_nan):
        """Test nan_to_mean with dataset containing no NaN values."""
        original_data = dataset_no_nan.hx.values.copy()
        ds_filled = nan_to_mean(dataset_no_nan.copy())

        # Data should remain unchanged
        assert np.allclose(ds_filled.hx.values, original_data)
        assert not np.any(np.isnan(ds_filled.hx.values))

    def test_nan_to_mean_all_nans(self, dataset_all_nan):
        """Test nan_to_mean with dataset containing all NaN values."""
        ds_filled = nan_to_mean(dataset_all_nan.copy())

        # Should replace with 0 (from np.nan_to_num of nanmean)
        assert np.allclose(ds_filled.hx.values, 0.0)

    def test_nan_to_mean_preserves_structure(self, multi_channel_dataset_with_nan):
        """Test that nan_to_mean preserves dataset structure."""
        ds_filled = nan_to_mean(multi_channel_dataset_with_nan.copy())

        # Check that coordinates are preserved
        assert np.allclose(
            ds_filled.time.values, multi_channel_dataset_with_nan.time.values
        )

        # Check that channels are preserved
        assert set(ds_filled.data_vars) == set(multi_channel_dataset_with_nan.data_vars)

    def test_nan_to_mean_single_nan_at_edges(self, subtests):
        """Test nan_to_mean with NaN at beginning and end."""
        times = np.array([0, 1, 2, 3, 4])

        test_cases = [
            (
                "nan_at_start",
                np.array([np.nan, 2.0, 3.0, 4.0, 5.0]),
                np.array([3.5, 2.0, 3.0, 4.0, 5.0]),
            ),
            (
                "nan_at_end",
                np.array([1.0, 2.0, 3.0, 4.0, np.nan]),
                np.array([1.0, 2.0, 3.0, 4.0, 2.5]),
            ),
            (
                "nan_at_both",
                np.array([np.nan, 2.0, 3.0, 4.0, np.nan]),
                np.array([3.0, 2.0, 3.0, 4.0, 3.0]),
            ),
        ]

        for name, data, expected in test_cases:
            with subtests.test(case=name):
                ds = xr.Dataset({"hx": ("time", data)}, coords={"time": times})
                ds_filled = nan_to_mean(ds.copy())
                assert np.allclose(ds_filled.hx.values, expected)


class TestHandleNanBasic:
    """Test basic handle_nan functionality."""

    def test_handle_nan_basic(self, extended_times):
        """Test basic functionality of handle_nan with NaN values."""
        data_x = np.array([1.0, np.nan, 3.0, 4.0, 5.0])
        data_y = np.array([1.0, 2.0, np.nan, 4.0, 5.0])

        X = xr.Dataset({"hx": ("time", data_x)}, coords={"time": extended_times})
        Y = xr.Dataset({"ex": ("time", data_y)}, coords={"time": extended_times})

        # Test with X and Y only
        X_clean, Y_clean, _ = handle_nan(X, Y, None, drop_dim="time")

        # Check that NaN values were dropped
        assert len(X_clean.time) == 3
        assert len(Y_clean.time) == 3
        assert not np.any(np.isnan(X_clean.hx.values))
        assert not np.any(np.isnan(Y_clean.ex.values))

        # Check that correct values remain
        expected_times = np.array([0, 3, 4])
        assert np.allclose(X_clean.time.values, expected_times)
        assert np.allclose(Y_clean.time.values, expected_times)

    def test_handle_nan_x_only(self):
        """Test handle_nan with only X dataset (Y empty, RR None)."""
        times = np.array([0, 1, 2, 3])
        data_x = np.array([1.0, np.nan, 3.0, 4.0])

        X = xr.Dataset({"hx": ("time", data_x)}, coords={"time": times})
        # Empty dataset with matching time coordinate
        Y = xr.Dataset(coords={"time": times})

        X_clean, Y_clean, RR_clean = handle_nan(X, Y, None, drop_dim="time")

        # Check that NaN was dropped from X
        assert len(X_clean.time) == 3
        assert not np.any(np.isnan(X_clean.hx.values))

        # Y and RR should be empty datasets
        assert len(Y_clean.data_vars) == 0
        assert len(RR_clean.data_vars) == 0

    def test_handle_nan_no_nans(self):
        """Test handle_nan with datasets containing no NaN values."""
        times = np.array([0, 1, 2, 3])
        data_x = np.array([1.0, 2.0, 3.0, 4.0])
        data_y = np.array([1.0, 2.0, 3.0, 4.0])

        X = xr.Dataset({"hx": ("time", data_x)}, coords={"time": times})
        Y = xr.Dataset({"ex": ("time", data_y)}, coords={"time": times})

        X_clean, Y_clean, _ = handle_nan(X, Y, None, drop_dim="time")

        # All data should be preserved
        assert len(X_clean.time) == 4
        assert len(Y_clean.time) == 4
        assert np.allclose(X_clean.hx.values, data_x)
        assert np.allclose(Y_clean.ex.values, data_y)


class TestHandleNanRemoteReference:
    """Test handle_nan with remote reference data."""

    def test_handle_nan_with_remote_reference(self):
        """Test handle_nan with remote reference data."""
        times = np.array([0, 1, 2, 3])
        data_x = np.array([1.0, np.nan, 3.0, 4.0])
        data_y = np.array([1.0, 2.0, 3.0, 4.0])
        data_rr = np.array([1.0, 2.0, np.nan, 4.0])

        X = xr.Dataset({"hx": ("time", data_x)}, coords={"time": times})
        Y = xr.Dataset({"ex": ("time", data_y)}, coords={"time": times})
        RR = xr.Dataset({"hx": ("time", data_rr)}, coords={"time": times})

        # Test with all datasets
        X_clean, Y_clean, RR_clean = handle_nan(X, Y, RR, drop_dim="time")

        # Check that NaN values were dropped
        assert len(X_clean.time) == 2
        assert len(Y_clean.time) == 2
        assert len(RR_clean.time) == 2
        assert not np.any(np.isnan(X_clean.hx.values))
        assert not np.any(np.isnan(Y_clean.ex.values))
        assert not np.any(np.isnan(RR_clean.hx.values))

        # Check that the values are correct
        expected_times = np.array([0, 3])
        assert np.allclose(X_clean.time.values, expected_times)
        assert np.allclose(Y_clean.time.values, expected_times)
        assert np.allclose(RR_clean.time.values, expected_times)
        assert np.allclose(X_clean.hx.values, np.array([1.0, 4.0]))
        assert np.allclose(Y_clean.ex.values, np.array([1.0, 4.0]))
        assert np.allclose(RR_clean.hx.values, np.array([1.0, 4.0]))

    def test_handle_nan_remote_reference_only(self):
        """Test handle_nan with only remote reference having NaN."""
        times = np.array([0, 1, 2, 3])
        data_x = np.array([1.0, 2.0, 3.0, 4.0])
        data_y = np.array([1.0, 2.0, 3.0, 4.0])
        data_rr = np.array([1.0, np.nan, 3.0, 4.0])

        X = xr.Dataset({"hx": ("time", data_x)}, coords={"time": times})
        Y = xr.Dataset({"ex": ("time", data_y)}, coords={"time": times})
        RR = xr.Dataset({"hy": ("time", data_rr)}, coords={"time": times})

        X_clean, Y_clean, RR_clean = handle_nan(X, Y, RR, drop_dim="time")

        # Only time index 1 should be dropped
        assert len(X_clean.time) == 3
        assert len(Y_clean.time) == 3
        assert len(RR_clean.time) == 3

        expected_times = np.array([0, 2, 3])
        assert np.allclose(X_clean.time.values, expected_times)

    def test_handle_nan_channel_name_preservation(self):
        """Test that channel names are preserved correctly with RR."""
        times = np.array([0, 1, 2])
        data = np.array([1.0, 2.0, 3.0])

        X = xr.Dataset({"hx": ("time", data)}, coords={"time": times})
        Y = xr.Dataset({"ex": ("time", data)}, coords={"time": times})
        RR = xr.Dataset(
            {"hx": ("time", data), "hy": ("time", data)}, coords={"time": times}
        )

        X_clean, Y_clean, RR_clean = handle_nan(X, Y, RR, drop_dim="time")

        # Check channel names
        assert "hx" in X_clean.data_vars
        assert "ex" in Y_clean.data_vars
        assert "hx" in RR_clean.data_vars
        assert "hy" in RR_clean.data_vars

        # RR channels should not have "remote_" prefix in output
        assert "remote_hx" not in RR_clean.data_vars


class TestHandleNanTimeMismatch:
    """Test handle_nan with time coordinate mismatches."""

    def test_handle_nan_time_mismatch(self):
        """Test handle_nan with time coordinate mismatches."""
        times_x = np.array([0, 1, 2, 3])
        times_rr = times_x + 0.1  # Small offset
        data_x = np.array([1.0, 2.0, 3.0, 4.0])
        data_rr = np.array([1.0, 2.0, 3.0, 4.0])

        X = xr.Dataset({"hx": ("time", data_x)}, coords={"time": times_x})
        RR = xr.Dataset({"hx": ("time", data_rr)}, coords={"time": times_rr})

        # Test handling of time mismatch
        X_clean, _, RR_clean = handle_nan(X, None, RR, drop_dim="time")

        # Check that data was preserved despite time mismatch
        assert len(X_clean.time) == 4
        assert "hx" in RR_clean.data_vars
        assert np.allclose(RR_clean.hx.values, data_rr)

        # Check that the time values match X's time values
        assert np.allclose(RR_clean.time.values, X_clean.time.values)

    def test_handle_nan_partial_time_mismatch(self):
        """Test handle_nan when only some time coordinates mismatch."""
        times_x = np.array([0.0, 1.0, 2.0, 3.0])
        times_rr = np.array([0.0, 1.0, 2.0001, 3.0])  # Slight mismatch at index 2
        data_x = np.array([1.0, 2.0, 3.0, 4.0])
        data_rr = np.array([1.0, 2.0, 3.0, 4.0])

        X = xr.Dataset({"hx": ("time", data_x)}, coords={"time": times_x})
        RR = xr.Dataset({"hy": ("time", data_rr)}, coords={"time": times_rr})

        # Should handle this with left join
        X_clean, _, RR_clean = handle_nan(X, None, RR, drop_dim="time")

        assert len(X_clean.time) == 4
        assert len(RR_clean.time) == 4


class TestTimeAxisMatch:
    """Test time_axis_match function."""

    def test_time_axis_match_exact(self):
        """Test time_axis_match when all axes match exactly."""
        times = np.array([0, 1, 2, 3])
        data = np.array([1.0, 2.0, 3.0, 4.0])

        X = xr.Dataset({"hx": ("time", data)}, coords={"time": times})
        Y = xr.Dataset({"ex": ("time", data)}, coords={"time": times})
        RR = xr.Dataset({"hy": ("time", data)}, coords={"time": times})

        assert time_axis_match(X, Y, RR) is True

    def test_time_axis_match_xy_only(self):
        """Test time_axis_match with only X and Y."""
        times = np.array([0, 1, 2, 3])
        data = np.array([1.0, 2.0, 3.0, 4.0])

        X = xr.Dataset({"hx": ("time", data)}, coords={"time": times})
        Y = xr.Dataset({"ex": ("time", data)}, coords={"time": times})

        assert time_axis_match(X, Y, None) is True

    def test_time_axis_match_x_rr_only(self):
        """Test time_axis_match with only X and RR."""
        times = np.array([0, 1, 2, 3])
        data = np.array([1.0, 2.0, 3.0, 4.0])

        X = xr.Dataset({"hx": ("time", data)}, coords={"time": times})
        RR = xr.Dataset({"hy": ("time", data)}, coords={"time": times})

        assert time_axis_match(X, None, RR) is True

    def test_time_axis_match_mismatch(self):
        """Test time_axis_match when axes do not match."""
        times_x = np.array([0, 1, 2, 3])
        times_rr = np.array([0, 1, 2, 4])  # Different last value
        data = np.array([1.0, 2.0, 3.0, 4.0])

        X = xr.Dataset({"hx": ("time", data)}, coords={"time": times_x})
        RR = xr.Dataset({"hy": ("time", data)}, coords={"time": times_rr})

        assert time_axis_match(X, None, RR) is False

    def test_time_axis_match_different_lengths(self):
        """Test time_axis_match with different length time axes."""
        times_x = np.array([0, 1, 2, 3])
        times_y = np.array([0, 1, 2])

        X = xr.Dataset(
            {"hx": ("time", np.array([1.0, 2.0, 3.0, 4.0]))}, coords={"time": times_x}
        )
        Y = xr.Dataset(
            {"ex": ("time", np.array([1.0, 2.0, 3.0]))}, coords={"time": times_y}
        )
        RR = xr.Dataset(
            {"hy": ("time", np.array([1.0, 2.0, 3.0, 4.0]))}, coords={"time": times_x}
        )

        # Use RR instead of None to avoid AttributeError
        assert time_axis_match(X, Y, RR) is False

    def test_time_axis_match_float_precision(self):
        """Test time_axis_match with floating point precision issues."""
        times_x = np.array([0.0, 0.1, 0.2, 0.3])
        times_rr = times_x + 1e-10  # Very small difference
        data = np.array([1.0, 2.0, 3.0, 4.0])

        X = xr.Dataset({"hx": ("time", data)}, coords={"time": times_x})
        RR = xr.Dataset({"hy": ("time", data)}, coords={"time": times_rr})

        # Should not match due to precision difference
        assert time_axis_match(X, None, RR) is False


class TestHandleNanMultipleChannels:
    """Test handle_nan with multiple channels in each dataset."""

    def test_handle_nan_multiple_channels_x_y(self):
        """Test handle_nan with multiple channels in X and Y."""
        times = np.array([0, 1, 2, 3])
        data_hx = np.array([1.0, np.nan, 3.0, 4.0])
        data_hy = np.array([1.0, 2.0, np.nan, 4.0])
        data_ex = np.array([np.nan, 2.0, 3.0, 4.0])
        data_ey = np.array([1.0, 2.0, 3.0, 4.0])

        X = xr.Dataset(
            {
                "hx": ("time", data_hx),
                "hy": ("time", data_hy),
            },
            coords={"time": times},
        )

        Y = xr.Dataset(
            {
                "ex": ("time", data_ex),
                "ey": ("time", data_ey),
            },
            coords={"time": times},
        )

        X_clean, Y_clean, _ = handle_nan(X, Y, None, drop_dim="time")

        # Only time index 3 has no NaN in any channel
        assert len(X_clean.time) == 1
        assert len(Y_clean.time) == 1
        assert X_clean.time.values[0] == 3

    def test_handle_nan_preserves_all_channels(self):
        """Test that all channels are preserved after NaN handling."""
        times = np.array([0, 1, 2])
        data = np.array([1.0, 2.0, 3.0])

        X = xr.Dataset(
            {
                "hx": ("time", data),
                "hy": ("time", data),
                "hz": ("time", data),
            },
            coords={"time": times},
        )

        Y = xr.Dataset(
            {
                "ex": ("time", data),
                "ey": ("time", data),
            },
            coords={"time": times},
        )

        X_clean, Y_clean, _ = handle_nan(X, Y, None, drop_dim="time")

        # All channels should be preserved
        assert set(X_clean.data_vars) == {"hx", "hy", "hz"}
        assert set(Y_clean.data_vars) == {"ex", "ey"}


class TestHandleNanEdgeCases:
    """Test edge cases for handle_nan."""

    def test_handle_nan_empty_dataset(self):
        """Test handle_nan with empty Y and RR."""
        times = np.array([0, 1, 2, 3])
        data = np.array([1.0, 2.0, 3.0, 4.0])

        X = xr.Dataset({"hx": ("time", data)}, coords={"time": times})
        # Empty dataset with matching time coordinate
        Y = xr.Dataset(coords={"time": times})

        X_clean, Y_clean, RR_clean = handle_nan(X, Y, None, drop_dim="time")

        # X should be unchanged
        assert len(X_clean.time) == 4
        assert np.allclose(X_clean.hx.values, data)

        # Y and RR should be empty
        assert len(Y_clean.data_vars) == 0
        assert len(RR_clean.data_vars) == 0

    def test_handle_nan_all_nans_dropped(self):
        """Test handle_nan when all rows have at least one NaN."""
        times = np.array([0, 1, 2])
        data_x = np.array([np.nan, 2.0, 3.0])
        data_y = np.array([1.0, np.nan, 3.0])
        data_rr = np.array([1.0, 2.0, np.nan])

        X = xr.Dataset({"hx": ("time", data_x)}, coords={"time": times})
        Y = xr.Dataset({"ex": ("time", data_y)}, coords={"time": times})
        RR = xr.Dataset({"hy": ("time", data_rr)}, coords={"time": times})

        X_clean, Y_clean, RR_clean = handle_nan(X, Y, RR, drop_dim="time")

        # No rows should remain
        assert len(X_clean.time) == 0
        assert len(Y_clean.time) == 0
        assert len(RR_clean.time) == 0

    def test_handle_nan_different_drop_dim(self):
        """Test handle_nan still works when drop_dim is specified (even though time_axis_match assumes 'time')."""
        # Note: time_axis_match function assumes 'time' dimension exists, so we use 'time' here
        # but test that drop_dim parameter is respected
        times = np.array([0, 1, 2, 3])
        data_x = np.array([1.0, np.nan, 3.0, 4.0])
        data_y = np.array([1.0, 2.0, 3.0, 4.0])

        X = xr.Dataset({"hx": ("time", data_x)}, coords={"time": times})
        Y = xr.Dataset({"ex": ("time", data_y)}, coords={"time": times})

        X_clean, Y_clean, _ = handle_nan(X, Y, None, drop_dim="time")

        # NaN at index 1 should be dropped
        assert len(X_clean.time) == 3
        assert len(Y_clean.time) == 3

        expected_times = np.array([0, 2, 3])
        assert np.allclose(X_clean.time.values, expected_times)


class TestHandleNanDataIntegrity:
    """Test that handle_nan preserves data integrity."""

    def test_handle_nan_values_correctness(self):
        """Test that correct values are preserved after dropping NaNs."""
        times = np.array([0, 1, 2, 3, 4])
        data_x = np.array([10.0, np.nan, 30.0, np.nan, 50.0])
        data_y = np.array([100.0, 200.0, np.nan, 400.0, 500.0])

        X = xr.Dataset({"hx": ("time", data_x)}, coords={"time": times})
        Y = xr.Dataset({"ex": ("time", data_y)}, coords={"time": times})

        X_clean, Y_clean, _ = handle_nan(X, Y, None, drop_dim="time")

        # Only times 0 and 4 have no NaN in any channel
        expected_times = np.array([0, 4])
        expected_x = np.array([10.0, 50.0])
        expected_y = np.array([100.0, 500.0])

        assert np.allclose(X_clean.time.values, expected_times)
        assert np.allclose(X_clean.hx.values, expected_x)
        assert np.allclose(Y_clean.ex.values, expected_y)

    def test_handle_nan_original_unchanged(self):
        """Test that original datasets are not modified by handle_nan."""
        times = np.array([0, 1, 2, 3])
        data_x = np.array([1.0, np.nan, 3.0, 4.0])
        data_y = np.array([1.0, 2.0, 3.0, 4.0])

        X = xr.Dataset({"hx": ("time", data_x)}, coords={"time": times})
        Y = xr.Dataset({"ex": ("time", data_y)}, coords={"time": times})

        # Store original values
        original_x_len = len(X.time)
        original_y_len = len(Y.time)

        # Call handle_nan
        X_clean, Y_clean, _ = handle_nan(X, Y, None, drop_dim="time")

        # Original datasets should be unchanged
        assert len(X.time) == original_x_len
        assert len(Y.time) == original_y_len
        assert np.isnan(X.hx.values[1])  # NaN still present



"""
This module contains helper methods that are used during transfer function processing.

Development Notes:
 Note #1: repeatedly applying edf_weights seems to have no effect at all.
 tested 20240118 and found that test_compare in synthetic passed whether this was commented
 or not.  TODO confirm this is a one-and-done add doc about why this is so.

"""

from aurora.time_series.frequency_band_helpers import get_band_for_tf_estimate
from aurora.time_series.xarray_helpers import handle_nan
from aurora.transfer_function.regression.base import RegressionEstimator
from aurora.transfer_function.regression.iter_control import IterControl
from aurora.transfer_function.regression.RME import RME
from aurora.transfer_function.regression.RME_RR import RME_RR

from aurora.transfer_function.weights.edf_weights import (
    effective_degrees_of_freedom_weights,
)
from mt_metadata.processing.aurora.decimation_level import (
    DecimationLevel as AuroraDecimationLevel,
)
from loguru import logger
from typing import Literal, Optional, Union

import numpy as np
import xarray as xr

ESTIMATOR_LIBRARY = {"OLS": RegressionEstimator, "RME": RME, "RME_RR": RME_RR}
SUPPORTED_REGRESSION_ESTIMATOR = Literal["OLS", "RME", "RME_RR"]


def get_estimator_class(
    estimation_engine: SUPPORTED_REGRESSION_ESTIMATOR,
) -> RegressionEstimator:
    """

    Parameters
    ----------
    estimation_engine: Literal["OLS", "RME", "RME_RR"]
        One of the keys in the ESTIMATOR_LIBRARY, designates the method that will be
        used to estimate the transfer function

    Returns
    -------
    estimator_class: aurora.transfer_function.regression.base.RegressionEstimator
        The class that will do the TF estimation
    """
    try:
        estimator_class = ESTIMATOR_LIBRARY[estimation_engine]
    except KeyError:
        logger.error(f"processing_scheme {estimation_engine} not supported")
        logger.error(
            f"processing_scheme must be one of {list(ESTIMATOR_LIBRARY.keys())}"
        )
        raise Exception
    return estimator_class


def set_up_iter_control(config: AuroraDecimationLevel):
    """
    Initializes an IterControl object based on values in the processing config.

    Development Notes:
    TODO: Review: maybe better to just make this the __init__ method of the IterControl object, iter_control = IterControl(config)


    Parameters
    ----------
    config: AuroraDecimationLevel
        metadata about the decimation level processing.

    Returns
    -------
    iter_control: aurora.transfer_function.regression.iter_control.IterControl
        Object with parameters about iteration control in regression
    """
    if config.estimator.engine in ["RME", "RME_RR"]:
        iter_control = IterControl(
            max_number_of_iterations=config.regression.max_iterations,
            max_number_of_redescending_iterations=config.regression.max_redescending_iterations,
            r0=config.regression.r0,
            u0=config.regression.u0,
            tolerance=config.regression.tolerance,
        )
    elif config.estimator.engine in [
        "OLS",
    ]:
        iter_control = None

    return iter_control


def drop_nans(X: xr.Dataset, Y: xr.Dataset, RR: Union[xr.Dataset, None]) -> tuple:
    """
    Drops any observation where any variable in X, Y, or RR is NaN.
    """
    import numpy as np

    def get_obs_mask(ds):
        """
        Generate a boolean mask indicating which 'observation' entries are finite across all data variables in an xarray Dataset.

        This function iterates over all data variables in the provided xarray Dataset `ds`, checks for finite values (i.e., not NaN or infinite)
        along all axes except the 'observation' axis, and combines the results to produce a single boolean mask. The resulting mask is True
        for each 'observation' index where all data variables are finite, and False otherwise.


        Parameters
        ds : xarray.Dataset
            The input dataset containing data variables with an 'observation' dimension.

        Returns
        numpy.ndarray
            A boolean array with shape matching the 'observation' dimension, where True indicates all data variables are finite
        """
        mask = None
        for v in ds.data_vars.values():
            # Reduce all axes except 'observation'
            axes = tuple(i for i, d in enumerate(v.dims) if d != "observation")
            this_mask = np.isfinite(v)
            if axes:
                this_mask = this_mask.all(axis=axes)
            mask = this_mask if mask is None else mask & this_mask
        return mask

    mask = get_obs_mask(X)
    mask = mask & get_obs_mask(Y)
    if RR is not None:
        mask = mask & get_obs_mask(RR)

    X = X.isel(observation=mask)
    Y = Y.isel(observation=mask)
    if RR is not None:
        RR = RR.isel(observation=mask)
    return X, Y, RR


def stack_fcs(X, Y, RR):
    """
    Reshape 2D arrays of frequency and time to 1D.

    Notes: When the data for a frequency band are extracted from the Spectrogram, each
    channel is a 2D array, one axis is time (the time of the window that was FFT-ed) and the
    other axis is frequency. However if we make no distinction between the harmonics
    (bins) within a band in regression, then all the FCs for each channel can be
    put into a 1D array.  This method performs that reshaping (ravelling) operation.
    It is not important how we unravel the FCs but it is important that the same indexing
    scheme is used for X, Y and RR.

    TODO: Consider this take a list and return a list rather than X,Y,RR

    TODO: Consider decorate this with @dataset_or_dataarray

    Parameters
    ----------
    X: xarray.core.dataset.Dataset
    Y: xarray.core.dataset.Dataset
    RR: xarray.core.dataset.Dataset or None

    Returns
    -------
    X, Y, RR: Same as input but with stacked time and frequency dimensions
    """
    X = X.stack(observation=("frequency", "time"))
    Y = Y.stack(observation=("frequency", "time"))
    if RR is not None:
        RR = RR.stack(observation=("frequency", "time"))
    return X, Y, RR


def apply_weights(
    X: xr.Dataset,
    Y: xr.Dataset,
    RR: xr.Dataset,
    W: np.ndarray,
    segment: bool = False,
    dropna: bool = False,
) -> tuple:
    """
    Applies data weights (W) to each of X, Y, RR.
    If weight is zero, we set to nan and optionally dropna.

    Parameters
    ----------
    X: xarray.core.dataset.Dataset
    Y: xarray.core.dataset.Dataset
    RR: xarray.core.dataset.Dataset or None
    W: numpy array
        The Weights to apply to the data
    segment: bool
        If True the weights may need to be reshaped.
    dropna: bool
        Whether or not to drop zero-weighted data.  If true, we drop the nans.

    Returns
    -------
    X, Y, RR: tuple
        Same as input but with weights applied and (optionally) nan dropped.
    """
    W[W == 0] = np.nan
    if segment:
        W = np.atleast_2d(W).T
    X *= W
    Y *= W
    if RR is not None:
        RR *= W

    if dropna:
        X, Y, RR = drop_nans(X, Y, RR)

    return X, Y, RR


def process_transfer_functions(
    dec_level_config: AuroraDecimationLevel,
    local_stft_obj: xr.Dataset,
    remote_stft_obj: xr.Dataset,
    transfer_function_obj,
):
    """
    This is the main tf_processing method.  It is based on the Matlab legacy code TTFestBand.m.

    Note #1: Although it is advantageous to execute the regression channel-by-channel
    vs. all-at-once, we need to keep the all-at-once to get residual covariances (see aurora issue #87)

    TODO: Consider push the nan-handling into the band extraction as a kwarg.

    Parameters
    ----------
    dec_level_config: AuroraDecimationLevel
        Processing parameters for the active decimation level.
    local_stft_obj: xarray.core.dataset.Dataset
    remote_stft_obj: xarray.core.dataset.Dataset or None
    transfer_function_obj: aurora.transfer_function.TTFZ.TTFZ
        The transfer function container ready to receive values in this method.

    Returns
    -------
    transfer_function_obj: aurora.transfer_function.TTFZ.TTFZ
    """
    estimator_class: RegressionEstimator = get_estimator_class(
        dec_level_config.estimator.engine
    )
    iter_control = set_up_iter_control(dec_level_config)
    for band in transfer_function_obj.frequency_bands.bands():

        X, Y, RR = get_band_for_tf_estimate(
            band, dec_level_config, local_stft_obj, remote_stft_obj
        )

        # Reshape to 2d
        X, Y, RR = stack_fcs(X, Y, RR)

        # Should only be needed if weights were applied
        X, Y, RR = drop_nans(X, Y, RR)

        W = effective_degrees_of_freedom_weights(X, RR, edf_obj=None)
        X, Y, RR = apply_weights(X, Y, RR, W, segment=False, dropna=True)

        if dec_level_config.estimator.estimate_per_channel:
            for ch in dec_level_config.output_channels:

                Y_ch = Y[ch].to_dataset()  # keep as a dataset, maybe not needed

                X_, Y_, RR_ = handle_nan(X, Y_ch, RR, drop_dim="observation")

                # see note #1
                # if RR is not None:
                #     W = effective_degrees_of_freedom_weights(X_, RR_, edf_obj=None)
                #     X_, Y_, RR_ = apply_weights(X_, Y_, RR_, W, segment=False)

                regression_estimator = estimator_class(
                    X=X_, Y=Y_, Z=RR_, iter_control=iter_control
                )
                regression_estimator.estimate()
                transfer_function_obj.set_tf(regression_estimator, band.center_period)
        else:
            X, Y, RR = handle_nan(X, Y, RR, drop_dim="observation")
            regression_estimator = estimator_class(
                X=X, Y=Y, Z=RR, iter_control=iter_control
            )
            regression_estimator.estimate()
            transfer_function_obj.set_tf(regression_estimator, band.center_period)

    return transfer_function_obj


def process_transfer_functions_with_weights(
    dec_level_config: AuroraDecimationLevel,
    local_stft_obj: xr.Dataset,
    remote_stft_obj: xr.Dataset,
    transfer_function_obj,
):
    """
    This is version of process_transfer_functions applies weights to the data.

    Development Notes:
    Note #1: This is only for per-channel estimation, so it does not support the
    dec_level_config.estimator.estimate_per_channel = False
    Note #2: This was adapted from the process_transfer_functions method but the core loop
    is inverted to loop over channels first, then bands.

    Parameters
    ----------
    dec_level_config: AuroraDecimationLevel
        Processing parameters for the active decimation level.
    local_stft_obj: xarray.core.dataset.Dataset
    remote_stft_obj: xarray.core.dataset.Dataset or None
    transfer_function_obj: aurora.transfer_function.TTFZ.TTFZ
        The transfer function container ready to receive values in this method.

    Returns
    -------
    transfer_function_obj: aurora.transfer_function.TTFZ.TTFZ
    """
    if not dec_level_config.estimator.estimate_per_channel:
        msg = (
            "process_transfer_functions_with_weights is only for per-channel estimation"
        )
        logger.error(msg)
        raise ValueError(msg)

    estimator_class: RegressionEstimator = get_estimator_class(
        dec_level_config.estimator.engine
    )
    iter_control = set_up_iter_control(dec_level_config)
    for ch in dec_level_config.output_channels:

        # check if there are channel weights for this channel
        weights = None
        for chws in dec_level_config.channel_weight_specs:
            if ch in chws.output_channels:
                weights = chws.weights

        for band in transfer_function_obj.frequency_bands.bands():

            X, Y, RR = get_band_for_tf_estimate(
                band, dec_level_config, local_stft_obj, remote_stft_obj
            )
            Y_ch = Y[ch].to_dataset()  # keep as a dataset, maybe not needed

            # extract the weights for this band
            if weights is not None:
                # TODO: Investigate best way to extract the weights for band
                #  This may involve finding the nearest frequency bin to the band center period
                #  and then applying the weights for that bin, or some tapered region around it.
                #  For now, we will just use the mean of the weights for the band.
                #  This is a temporary solution and should be replaced with a more robust method.
                # band_weights = chws.get_weights_for_band(band)
                band_weights = weights.mean(axis=1)  # chws.get_weights_for_band(band)

                apply_weights(
                    X, Y_ch, RR, band_weights.squeeze(), segment=True, dropna=False
                )

            X, Y_ch, RR = stack_fcs(X, Y_ch, RR)  # Reshape to 2d

            # Should only be needed if weights were applied
            X, Y_ch, RR = drop_nans(X, Y_ch, RR)

            W = effective_degrees_of_freedom_weights(X, RR, edf_obj=None)

            X, Y_ch, RR = apply_weights(X, Y_ch, RR, W, segment=False, dropna=True)
            X_, Y_, RR_ = handle_nan(X, Y_ch, RR, drop_dim="observation")
            regression_estimator = estimator_class(
                X=X_, Y=Y_, Z=RR_, iter_control=iter_control
            )
            regression_estimator.estimate()
            transfer_function_obj.set_tf(regression_estimator, band.center_period)

    return transfer_function_obj



"""
    In this module is a class that emulates the old EMTF Band Setup File

"""
import pathlib
from typing import Optional, Union

import numpy as np
import pandas as pd
from loguru import logger


class EMTFBandSetupFile:
    def __init__(
        self,
        filepath: Optional[Union[str, pathlib.Path, None]] = None,
        sample_rate: Optional[Union[float, None]] = None,
    ):
        """
        Constructor.

        Parameters
        ----------
        sample_rate: float
            this is the sampling rate of the un-decimated data,
            aka decimation level 1 in EMTF (and probably zero in the
            aurora nomenclature)
        """
        self.filepath = filepath
        self.df = None
        self._num_bands = None
        self.sample_rate = sample_rate
        if self.filepath:
            self.load()

    @property
    def num_bands(self) -> int:
        """return the number of bands in the band setup file"""
        return self._num_bands

    @property
    def num_decimation_levels(self) -> int:
        """Return the number of decimation levels"""
        num_dec = len(self.decimation_levels)
        return num_dec

    @property
    def decimation_levels(self) -> np.ndarray:
        """Return the decimation level names (they are integers) in an array"""
        dec_levels = self.df.decimation_level.unique()
        dec_levels.sort()
        return dec_levels

    def load(self, filepath: Optional[Union[str, pathlib.Path, None]] = None) -> None:
        """
        Loads an EMTF band setup file and casts it to a dataframe.
        - populates self.df with the EMTF band setup dataframe.

        Parameters
        ----------
        filepath: Optional[Union[str, pathlib.Path, None]]
            The path to the emtf band setup (.cfrg) file.

        """
        if filepath is None:
            filepath = self.filepath
        msg = f"loading band setup file {filepath}"
        logger.debug(msg)
        with open(str(filepath), "r") as f:
            num_bands = f.readline()
        self._num_bands = int(num_bands)
        f.close()
        df = pd.read_csv(
            filepath,
            skiprows=1,
            sep=r"\s+",
            names=["decimation_level", "lower_bound_index", "upper_bound_index"],
        )
        if len(df) != self.num_bands:
            msg = f"unexpected number of bounds read in from {filepath}"
            logger.exception(msg)
            raise Exception(msg)
        self.df = df

    def get_decimation_level(
        self, decimation_level: int, order: str = "ascending_frequency"
    ) -> pd.DataFrame:
        """
        Return a sub-dataframe with only the rows that correspond to the requested decimation level

        Parameters
        ----------
        decimation_level: int
            The id of the decimation level
        order: str

        Returns
        -------
        decimation_level_df: pd.DataFrame
            A sub-dataframe with only the rows that correspond to the requested decimation level
        """
        if self.df is None:
            self.load()
        decimation_level_df = self.df[self.df["decimation_level"] == decimation_level]
        if order == "ascending_frequency":
            decimation_level_df = decimation_level_df.sort_values(
                by="lower_bound_index"
            )
        else:
            msg = f"order of decimation dataframe {order} not supported -- returning default order"
            logger.warning(msg)
        return decimation_level_df

    def compute_band_edges(
        self, decimation_factors: list, num_samples_window: int
    ) -> dict:
        """
        Adds columns to df defining the upper and lower bounds of the frequency bands in Hz.

        Parameters
        ----------
        decimation_factors: list
            The decimation factors
        num_samples_window: int
            The window size

        Returns
        -------
        band_edges: dict
            Keys are decimation level ids and values are numpy arrays with shape num_bands x 2.
        """
        band_edges = {}
        lower_edges = pd.Series(index=self.df.index, dtype="float64")
        upper_edges = pd.Series(index=self.df.index, dtype="float64")
        if not self.sample_rate:
            msg = "cannot define frequencies if sample rate undefined"
            logger.error(msg)
            raise ValueError(msg)
        if len(decimation_factors) != self.num_decimation_levels:
            logger.info(
                "Number of decimation_factors must equal number of decimation levels"
            )
            logger.info(
                f"There are {len(decimation_factors)} decimation_factors but "
                f"{self.num_decimation_levels} decimation levels"
            )
        # add a similar check for num_samples_window
        for i_dec, decimation_level in enumerate(self.decimation_levels):
            indices = self.df.decimation_level == decimation_level

            # emtf decimation levels start at 1, but i_dec indexes them from 0
            downsample_factor = 1.0 * np.prod(decimation_factors[0 : i_dec + 1])
            sample_rate = self.sample_rate / downsample_factor
            delta_f = sample_rate / num_samples_window[i_dec]
            half_df = delta_f / 2.0
            lower_edges[indices] = self.df["lower_bound_index"] * delta_f - half_df
            upper_edges[indices] = self.df["upper_bound_index"] * delta_f + half_df
            band_edges[i_dec] = np.vstack(
                (lower_edges[indices].values, upper_edges[indices].values)
            ).T

        self.df["lower_edge"] = lower_edges
        self.df["upper_edge"] = upper_edges
        return band_edges

    # def to_band_averaging_scheme(self):
    #     """
    #     probably better to give band averaging scheme a "from emtf"
    #     method
    #     Returns
    #     -------
    #
    #     """
    #     pass



"""
    Staging area for functions we want to import directly from transfer_function.
"""

from .transfer_function_collection import TransferFunctionCollection

__all__ = [
    "TransferFunctionCollection",
]



"""
    This module contains functions that are associated with time series of Fourier coefficients
    TODO: Move these methods to mth5.processing.spectre.frequency_band_helpers
"""
from loguru import logger
from mt_metadata.processing.aurora import (
    DecimationLevel as AuroraDecimationLevel,
)
from mt_metadata.processing.aurora import Band
from mth5.timeseries.spectre.spectrogram import extract_band
from typing import Optional, Tuple
import xarray as xr


def get_band_for_tf_estimate(
    band: Band,
    dec_level_config: AuroraDecimationLevel,
    local_stft_obj: xr.Dataset,
    remote_stft_obj: Optional[xr.Dataset],
) -> Tuple[xr.Dataset, xr.Dataset, Optional[xr.Dataset]]:
    """
    Returns spectrograms X, Y, RR for harmonics within the given band

    Parameters
    ----------
    band : mt_metadata.processing.aurora.Band
        object with lower_bound and upper_bound to tell stft object which
        subarray to return
    config : AuroraDecimationLevel
        information about the input and output channels needed for TF
        estimation problem setup
    local_stft_obj : xarray.core.dataset.Dataset or None
        Time series of Fourier coefficients for the station whose TF is to be
        estimated
    remote_stft_obj : xarray.core.dataset.Dataset or None
        Time series of Fourier coefficients for the remote reference station

    Returns
    -------
    X, Y, RR : xarray.core.dataset.Dataset or None
        data structures as local_stft_object and remote_stft_object, but
        restricted only to input_channels, output_channels,
        reference_channels and also the frequency axes are restricted to
        being within the frequency band given as an input argument.
    """
    logger.info(
        f"Accessing band {band.center_period:.6f}s  ({1./band.center_period:.6f}Hz)"
    )
    band_dataset = extract_band(band, local_stft_obj)
    X = band_dataset[dec_level_config.input_channels]
    Y = band_dataset[dec_level_config.output_channels]
    check_time_axes_synched(X, Y)
    if dec_level_config.reference_channels:
        band_dataset = extract_band(band, remote_stft_obj)
        RR = band_dataset[dec_level_config.reference_channels]
        check_time_axes_synched(Y, RR)
    else:
        RR = None

    return X, Y, RR


def check_time_axes_synched(X, Y):
    """
    Utility function for checking that time axes agree.
    Raises ValueError if axes do not agree.

    It is critical that X, Y, RR have the same time axes for aurora processing.

    Parameters
    ----------
    X : xarray
    Y : xarray


    """
    if (X.time == Y.time).all():
        pass
    else:
        msg = "Time axes of arrays not identical"
        #  "NAN Handling could fail if X,Y dont share time axes"
        logger.warning(msg)
        raise ValueError(msg)
    return


def adjust_band_for_coherence_sorting(frequency_band, spectrogram, rule="min3"):
    """

    WIP: Intended to broaden band to allow more FCs for spectral features
    - used in coherence sorting and general feature extraction

    Parameters
    ----------
    frequency_band
    spectrogram: Spectrogram
    rule

    Returns
    -------

    """
    band = frequency_band.copy()
    if spectrogram.num_harmonics_in_band(band) == 1:
        logger.warning("Cant evaluate coherence with only 1 harmonic")
        logger.info(f"Widening band according to {rule} rule")
        if rule == "min3":
            band.frequency_min -= spectrogram.frequency_increment
            band.frequency_max += spectrogram.frequency_increment
        else:
            msg = f"Band adjustment rule {rule} not recognized"
            logger.error(msg)
            raise NotImplementedError(msg)
    return band


# TODO: move this to mth5.processing.spectre
def get_band_for_coherence_sorting(
    frequency_band,
    dec_level_config: AuroraDecimationLevel,
    local_stft_obj,
    remote_stft_obj,
    widening_rule="min3",
):
    """
    Just like get_band_for_tf_estimate, but here we enforce some rules so that the band is not one FC wide
    - it is possible that this method will get merged with get_band_for_tf_estimate
    - this is a placeholder until the appropriate rules are sorted out.

    Parameters
    ----------
    band : mt_metadata.processing.aurora.FrequencyBands
        object with lower_bound and upper_bound to tell stft object which
        subarray to return
    config : AuroraDecimationLevel
        information about the input and output channels needed for TF
        estimation problem setup
    local_stft_obj : xarray.core.dataset.Dataset or None
        Time series of Fourier coefficients for the station whose TF is to be
        estimated
    remote_stft_obj : xarray.core.dataset.Dataset or None
        Time series of Fourier coefficients for the remote reference station

    Returns
    -------
    X, Y, RR : xarray.core.dataset.Dataset or None
        data structures as local_stft_object and remote_stft_object, but
        restricted only to input_channels, output_channels,
        reference_channels and also the frequency axes are restricted to
        being within the frequency band given as an input argument.
    """
    from mth5.timeseries.spectre import Spectrogram

    band = frequency_band.copy()
    logger.info(
        f"Processing band {band.center_period:.6f}s  ({1./band.center_period:.6f}Hz)"
    )
    stft = Spectrogram(local_stft_obj)
    if stft.num_harmonics_in_band(band) == 1:
        logger.warning("Cant evaluate coherence with only 1 harmonic")
        logger.info(f"Widening band according to {widening_rule} rule")
        if widening_rule == "min3":
            band.frequency_min -= stft.frequency_increment
            band.frequency_max += stft.frequency_increment
        else:
            msg = f"Widening rule {widening_rule} not recognized"
            logger.error(msg)
            raise NotImplementedError(msg)
    # return band TODO: Make this return the band only, then get later
    # proceed as in
    x, y, rr = get_band_for_tf_estimate(
        band, dec_level_config, local_stft_obj, remote_stft_obj
    )
    return x, y, rr, band


def cross_spectra(X, Y):
    """WIP: Returns the cross power spectra between two arrays"""
    return X.conj() * Y



"""Pytest translation of test_processing.py

Runs several synthetic processing tests from config creation to tf_cls.
"""

import logging

import pytest

from aurora.test_utils.synthetic.processing_helpers import (
    process_synthetic_1,
    process_synthetic_1r2,
    process_synthetic_2,
)


@pytest.fixture(autouse=True)
def setup_logging():
    """Disable noisy matplotlib loggers."""
    logging.getLogger("matplotlib.font_manager").disabled = True
    logging.getLogger("matplotlib.ticker").disabled = True


@pytest.mark.skip(
    reason="mt_metadata pydantic branch has issue with provenance.archive.comments.value being None"
)
def test_no_crash_with_too_many_decimations(synthetic_test_paths):
    """Test processing with many decimation levels."""
    z_file_path = synthetic_test_paths.aurora_results_path.joinpath("syn1_tfk.zss")
    xml_file_name = synthetic_test_paths.aurora_results_path.joinpath("syn1_tfk.xml")
    tf_cls = process_synthetic_1(config_keyword="test1_tfk", z_file_path=z_file_path)
    tf_cls.write(fn=xml_file_name, file_type="emtfxml")
    tf_cls.write(
        fn=z_file_path.parent.joinpath(f"{z_file_path.stem}_from_tf.zss"),
        file_type="zss",
    )

    xml_file_name = synthetic_test_paths.aurora_results_path.joinpath("syn1r2_tfk.xml")
    tf_cls = process_synthetic_1r2(config_keyword="test1r2_tfk")
    tf_cls.write(fn=xml_file_name, file_type="emtfxml")


class TestSyntheticTest1Processing:
    """Tests for test1 synthetic processing - share processed TF across tests."""

    @pytest.fixture(scope="class")
    def processed_tf_test1(self, worker_safe_test1_h5):
        """Process test1 once and reuse across all tests in this class."""
        return process_synthetic_1(file_version="0.1.0", mth5_path=worker_safe_test1_h5)

    def test_can_output_tf_class_and_write_tf_xml(
        self, synthetic_test_paths, processed_tf_test1
    ):
        """Test basic TF processing and XML output."""
        xml_file_name = synthetic_test_paths.aurora_results_path.joinpath(
            "syn1_mth5-010.xml"
        )
        processed_tf_test1.write(fn=xml_file_name, file_type="emtfxml")

    def test_can_use_mth5_file_version_020(
        self, synthetic_test_paths, processed_tf_test1
    ):
        """Test processing with MTH5 file version 0.2.0."""
        file_version = "0.2.0"
        z_file_path = synthetic_test_paths.aurora_results_path.joinpath(
            f"syn1-{file_version}.zss"
        )
        xml_file_name = synthetic_test_paths.aurora_results_path.joinpath(
            f"syn1_mth5v{file_version}.xml"
        )
        processed_tf_test1.write(fn=xml_file_name, file_type="emtfxml")
        processed_tf_test1.write(
            fn=z_file_path.parent.joinpath(f"{z_file_path.stem}_from_tf.zss"),
            file_type="zss",
        )

    @pytest.fixture(scope="class")
    def processed_tf_scaled(self, worker_safe_test1_h5, synthetic_test_paths):
        """Process test1 with scale factors once and reuse."""
        z_file_path = synthetic_test_paths.aurora_results_path.joinpath(
            "syn1-scaled.zss"
        )
        return process_synthetic_1(
            z_file_path=z_file_path,
            test_scale_factor=True,
            mth5_path=worker_safe_test1_h5,
        )

    def test_can_use_scale_factor_dictionary(
        self, processed_tf_scaled, synthetic_test_paths
    ):
        """Test channel scale factors in mt_metadata processing class.

        Expected outputs are four .png:
        - xy_syn1.png: Shows expected 100 Ohm-m resistivity
        - xy_syn1-scaled.png: Overestimates by 4x for 300 Ohm-m resistivity
        - yx_syn1.png: Shows expected 100 Ohm-m resistivity
        - yx_syn1-scaled.png: Underestimates by 4x for 25 Ohm-m resistivity
        """
        z_file_path = synthetic_test_paths.aurora_results_path.joinpath(
            "syn1-scaled.zss"
        )
        processed_tf_scaled.write(
            fn=z_file_path.parent.joinpath(f"{z_file_path.stem}_from_tf.zss"),
            file_type="zss",
        )

    @pytest.fixture(scope="class")
    def processed_tf_simultaneous(self, worker_safe_test1_h5, synthetic_test_paths):
        """Process test1 with simultaneous regression once and reuse."""
        z_file_path = synthetic_test_paths.aurora_results_path.joinpath(
            "syn1_simultaneous_estimate.zss"
        )
        return process_synthetic_1(
            z_file_path=z_file_path,
            simultaneous_regression=True,
            mth5_path=worker_safe_test1_h5,
        )

    def test_simultaneous_regression(
        self, processed_tf_simultaneous, synthetic_test_paths
    ):
        """Test simultaneous regression processing."""
        xml_file_name = synthetic_test_paths.aurora_results_path.joinpath(
            "syn1_simultaneous_estimate.xml"
        )
        z_file_path = synthetic_test_paths.aurora_results_path.joinpath(
            "syn1_simultaneous_estimate.zss"
        )
        processed_tf_simultaneous.write(fn=xml_file_name, file_type="emtfxml")
        processed_tf_simultaneous.write(
            fn=z_file_path.parent.joinpath(f"{z_file_path.stem}_from_tf.zss"),
            file_type="zss",
        )


def test_can_use_channel_nomenclature(synthetic_test_paths, mth5_target_dir, worker_id):
    """Test processing with custom channel nomenclature.

    Note: This test creates its own MTH5 with specific nomenclature, so it cannot
    share fixtures with other tests.
    """
    from mth5.data.make_mth5_from_asc import create_test1_h5

    channel_nomenclature = "LEMI12"
    # Create MTH5 with specific nomenclature in worker-safe directory
    mth5_path = create_test1_h5(
        file_version="0.1.0",
        channel_nomenclature=channel_nomenclature,
        target_folder=mth5_target_dir,
    )

    z_file_path = synthetic_test_paths.aurora_results_path.joinpath(
        f"syn1-{channel_nomenclature}.zss"
    )
    tf_cls = process_synthetic_1(
        z_file_path=z_file_path,
        file_version="0.1.0",
        channel_nomenclature=channel_nomenclature,
        mth5_path=mth5_path,
    )
    xml_file_name = synthetic_test_paths.aurora_results_path.joinpath(
        f"syn1_mth5-0.1.0_{channel_nomenclature}.xml"
    )
    tf_cls.write(fn=xml_file_name, file_type="emtfxml")


class TestSyntheticTest2Processing:
    """Tests for test2 synthetic processing."""

    @pytest.fixture(scope="class")
    def processed_tf_test2(self, worker_safe_test2_h5):
        """Process test2 once and reuse."""
        return process_synthetic_2(force_make_mth5=True, mth5_path=worker_safe_test2_h5)

    def test_can_process_other_station(self, synthetic_test_paths, processed_tf_test2):
        """Test processing a different synthetic station."""
        xml_file_name = synthetic_test_paths.aurora_results_path.joinpath("syn2.xml")
        processed_tf_test2.write(fn=xml_file_name, file_type="emtfxml")


class TestRemoteReferenceProcessing:
    """Tests for remote reference processing."""

    @pytest.fixture(scope="class")
    def processed_tf_test12rr(self, worker_safe_test12rr_h5):
        """Process test12rr once and reuse."""
        return process_synthetic_1r2(
            channel_nomenclature="default", mth5_path=worker_safe_test12rr_h5
        )

    def test_can_process_remote_reference_data(
        self, synthetic_test_paths, processed_tf_test12rr
    ):
        """Test remote reference processing with default channel nomenclature."""
        xml_file_name = synthetic_test_paths.aurora_results_path.joinpath(
            "syn12rr_mth5-010.xml"
        )
        processed_tf_test12rr.write(fn=xml_file_name, file_type="emtfxml")


def test_can_process_remote_reference_data_with_channel_nomenclature(
    synthetic_test_paths,
    mth5_target_dir,
    worker_id,
):
    """Test remote reference processing with custom channel nomenclature.

    Note: This test creates its own MTH5 with specific nomenclature, so it cannot
    share fixtures with other tests.
    """
    from mth5.data.make_mth5_from_asc import create_test12rr_h5

    channel_nomenclature = "LEMI34"
    # Create MTH5 with specific nomenclature in worker-safe directory
    mth5_path = create_test12rr_h5(
        channel_nomenclature=channel_nomenclature,
        target_folder=mth5_target_dir,
    )

    tf_cls = process_synthetic_1r2(
        channel_nomenclature=channel_nomenclature, mth5_path=mth5_path
    )
    xml_file_name = synthetic_test_paths.aurora_results_path.joinpath(
        "syn12rr_mth5-010_LEMI34.xml"
    )
    tf_cls.write(fn=xml_file_name, file_type="emtfxml")



"""
Tests comparing aurora processing results against archived EMTF results.

Development Notes:
- In the early days of development these tests were useful to check that
the same results were obtained from processing an mth5 with both stations in it
and two separate mth5 files.  This could probably be made into a simpler test in mth5
that checks that the data are the same in the two files.
- This used to make a homebrew resistivity and phase plot for comparison between
archived emtf z-files, but has been replaced with mt_metadata methods.
- TODO: Check phases in these plots -- they are off by 180 so there may be a sign error
in the data, or maybe the emtf results are using a different convention.  Need to investigate.
- The comparison with the matlab emtf results uses a slighly different windowing method.

"""
import numpy as np
import pytest
from loguru import logger
from mth5.helpers import close_open_files
from mth5.processing import KernelDataset, RunSummary

from aurora.general_helper_functions import DATA_PATH
from aurora.pipelines.process_mth5 import process_mth5
from aurora.test_utils.synthetic.make_processing_configs import create_test_run_config
from aurora.transfer_function.compare import CompareTF


# Path to baseline EMTF results in source tree
BASELINE_EMTF_PATH = DATA_PATH.joinpath("synthetic", "emtf_results")


def aurora_vs_emtf(
    synthetic_test_paths,
    test_case_id,
    emtf_version,
    auxilliary_z_file,
    z_file_base,
    tfk_dataset,
    atol_phase=4.0,
    make_rho_phi_plot=True,
):
    """
    Compare aurora processing results against EMTF baseline.

    Parameters
    ----------
    synthetic_test_paths : SyntheticTestPaths
        Path fixture for test directories
    test_case_id: str
        one of ["test1", "test2r1"]. "test1" is single station, "test2r1" is remote reference
    emtf_version: str
        one of ["fortran", "matlab"]
    auxilliary_z_file: str or pathlib.Path
        points to a .zss, .zrr or .zmm that EMTF produced
    z_file_base: str
        z_file basename for aurora output
    tfk_dataset: aurora.transfer_function.kernel_dataset.KernelDataset
        Info about data to process
    make_rho_phi_plot: bool
    show_rho_phi_plot: bool
    use_subtitle: bool
    """
    AURORA_RESULTS_PATH = synthetic_test_paths.aurora_results_path

    processing_config = create_test_run_config(
        test_case_id, tfk_dataset, matlab_or_fortran=emtf_version
    )

    z_file_path = AURORA_RESULTS_PATH.joinpath(z_file_base)

    process_mth5(
        processing_config,
        tfk_dataset=tfk_dataset,
        z_file_path=z_file_path,
        return_collection=True,
    )
    comparator = CompareTF(tf_01=z_file_path, tf_02=auxilliary_z_file)
    result = comparator.compare_transfer_functions(atol_phase=4.0)
    assert np.isclose(result["impedance_ratio"]["Z_10"], 1.0, atol=1e-2)
    assert np.isclose(result["impedance_ratio"]["Z_01"], 1.0, atol=1e-2)
    assert result["impedance_phase_close"]
    if make_rho_phi_plot:
        comparator.plot_two_transfer_functions(
            save_plot_path=AURORA_RESULTS_PATH.joinpath(
                f"{test_case_id}_aurora_vs_emtf_{emtf_version}_tf_compare.png"
            ),
            rho_xy_ylims=(10.0, 1000.0),
            rho_yx_ylims=(10.0, 1000.0),
        )


@pytest.mark.slow
def test_pipeline_merged(synthetic_test_paths, subtests, worker_safe_test12rr_h5):
    """Test aurora vs EMTF comparison with merged mth5."""
    close_open_files()

    # Create merged mth5
    mth5_path = worker_safe_test12rr_h5
    mth5_paths = [mth5_path]

    run_summary = RunSummary()
    run_summary.from_mth5s(mth5_paths)

    # Test1 vs fortran
    with subtests.test(case="test1", version="fortran"):
        logger.info("Test1 vs fortran")
        tfk_dataset = KernelDataset()
        tfk_dataset.from_run_summary(run_summary, "test1")
        auxilliary_z_file = BASELINE_EMTF_PATH.joinpath("test1.zss")
        z_file_base = "test1_aurora_fortran.zss"
        aurora_vs_emtf(
            synthetic_test_paths,
            "test1",
            "fortran",
            auxilliary_z_file,
            z_file_base,
            tfk_dataset,
        )

    # Test1 vs matlab
    with subtests.test(case="test1", version="matlab"):
        logger.info("Test1 vs matlab")
        tfk_dataset = KernelDataset()
        tfk_dataset.from_run_summary(run_summary, "test1")
        auxilliary_z_file = BASELINE_EMTF_PATH.joinpath("test1.zss")
        z_file_base = "test1_aurora_matlab.zss"
        aurora_vs_emtf(
            synthetic_test_paths,
            "test1",
            "matlab",
            auxilliary_z_file,
            z_file_base,
            tfk_dataset,
        )

    # Test2r1 vs fortran
    with subtests.test(case="test2r1", version="fortran"):
        logger.info("Test2r1")
        tfk_dataset = KernelDataset()
        tfk_dataset.from_run_summary(run_summary, "test2", "test1")
        auxilliary_z_file = BASELINE_EMTF_PATH.joinpath("test2r1.zrr")
        z_file_base = "test2r1_aurora_fortran.zrr"
        aurora_vs_emtf(
            synthetic_test_paths,
            "test2r1",
            "fortran",
            auxilliary_z_file,
            z_file_base,
            tfk_dataset,
        )


@pytest.mark.slow
def test_pipeline_separate(
    synthetic_test_paths, subtests, worker_safe_test1_h5, worker_safe_test2_h5
):
    """Test aurora vs EMTF comparison with separate mth5 files."""
    close_open_files()

    # Create separate mth5 files
    mth5_path_1 = worker_safe_test1_h5
    mth5_path_2 = worker_safe_test2_h5
    mth5_paths = [mth5_path_1, mth5_path_2]

    run_summary = RunSummary()
    run_summary.from_mth5s(mth5_paths)

    # Test1 vs fortran
    with subtests.test(case="test1", version="fortran"):
        logger.info("Test1 vs fortran")
        tfk_dataset = KernelDataset()
        tfk_dataset.from_run_summary(run_summary, "test1")
        auxilliary_z_file = BASELINE_EMTF_PATH.joinpath("test1.zss")
        z_file_base = "test1_aurora_fortran.zss"
        aurora_vs_emtf(
            synthetic_test_paths,
            "test1",
            "fortran",
            auxilliary_z_file,
            z_file_base,
            tfk_dataset,
        )

    # Test1 vs matlab
    with subtests.test(case="test1", version="matlab"):
        logger.info("Test1 vs matlab")
        tfk_dataset = KernelDataset()
        tfk_dataset.from_run_summary(run_summary, "test1")
        auxilliary_z_file = BASELINE_EMTF_PATH.joinpath("test1.zss")
        z_file_base = "test1_aurora_matlab.zss"
        aurora_vs_emtf(
            synthetic_test_paths,
            "test1",
            "matlab",
            auxilliary_z_file,
            z_file_base,
            tfk_dataset,
        )

    # Test2r1 vs fortran
    with subtests.test(case="test2r1", version="fortran"):
        logger.info("Test2r1")
        tfk_dataset = KernelDataset()
        tfk_dataset.from_run_summary(run_summary, "test2", "test1")
        auxilliary_z_file = BASELINE_EMTF_PATH.joinpath("test2r1.zrr")
        z_file_base = "test2r1_aurora_fortran.zrr"
        aurora_vs_emtf(
            synthetic_test_paths,
            "test2r1",
            "fortran",
            auxilliary_z_file,
            z_file_base,
            tfk_dataset,
        )



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""

This module is concerned with windowing time series.


Development Notes:

The windowing scheme defines the chunking and chopping of the time series for
the Short Time Fourier Transform.  Often referred to as a "sliding window" or
a "striding window".  In its most basic form it is a taper with a rule to
say how far to advance at each stride (or step).

To generate an array of data-windows from a data series we only need the
two parameters window_length (L) and window_overlap (V).  The parameter
"window_advance" (L-V) can be used in lieu of overlap.  Sliding windows are
normally described terms of overlap but it is cleaner to code in terms of
advance.

Choices L and V are usually made with some knowledge of time series sample
rate, duration, and the frequency band of interest.  In aurora because this is used
to prep for STFT, L is typically a power of 2.

In general we will need one instance of this class per decimation level,
but in practice often leave the windowing scheme the same for each decimation level.

This class is a key part of the "gateway" to frequency domain, so it has been given
a sampling_rate attribute.  While sampling rate is a property of the data, and not
the windowing scheme per se, it is good for this class to be aware of the sampling
rate.

Future modifications could involve:
- binding this class with a time series.
- Making a subclass with only L, V, and then having an extension with sample_rate


When 2D arrays are generated how should we index them?
|    [[ 0  1  2]
|    [ 2  3  4]
|    [ 4  5  6]
|    [ 6  7  8]
|    [ 8  9 10]
|    [10 11 12]
|    [12 13 14]]

In this example the rows are indexing the individual windows ... and so they
should be associated with the time of each window.  Will need a standard for
this.  Obvious options are center_time of window and time_of_first
sample. I prefer time_of_first sample.  This can always be transformed to
center time or another standard later.  We can call this the "window time
axis".  The columns are indexing "steps of delta-t".  The actual times are
different for every row, so it would be best to use something like
[0, dt, 2*dt] for that axis to keep it general.  We can call this the
"within-window sample time axis"


TODO: Regarding the optional time_vector input to self.apply_sliding_window()
... this current implementation takes as input numpy array data.  We need to
also allow for an xarray to be implemented. In the simplest case we would
take an xarray in and extract its "time" axis as time vector

20210529
This class is going to be modified to only accept xarray as input data.
We can force any incoming numpy arrays to be either xr.DataArray or xr.Dataset.
Similarly, output will be only xr.DataArray or xr.Dataset
"""

import copy
import numpy as np
import xarray as xr

from aurora.time_series.apodization_window import ApodizationWindow
from aurora.time_series.windowed_time_series import WindowedTimeSeries
from aurora.time_series.window_helpers import available_number_of_windows_in_array
from aurora.time_series.window_helpers import SLIDING_WINDOW_FUNCTIONS
from mt_metadata.processing.aurora.decimation_level import (
    DecimationLevel as AuroraDecimationLevel,
)
from mt_metadata.processing.window import get_fft_harmonics

from loguru import logger
from typing import Optional, Union


class WindowingScheme(ApodizationWindow):
    """

    Development notes
    20210415: Casting window length, overlap, advance, etc. in terms of number
    of samples ("taps", "points") here.  May allow defining these in terms
    of percent, duration in seconds etc. in future.

    Note: Technically, sample_rate is a property of the data, and not of the
    window ... but once the window is applied to data, the sample rate is defined.
    Sample rate is defined here because this window will operate on time series
    with a defined time axis.

    TODO: Add dtypes and docstrings.  Is it possible to replace kwargs with Optional arguments?
    """

    def __init__(self, **kwargs):
        """
        Constructor

        Parameters
        ----------
        kwargs
        """
        super(WindowingScheme, self).__init__(**kwargs)
        self.num_samples_overlap = kwargs.get(
            "num_samples_overlap",
            None,  # TODO: make this 75% of num_samples_window by default
        )
        self.striding_function_label = kwargs.get("striding_function_label", "crude")
        self._left_hand_window_edge_indices = None
        self.sample_rate = kwargs.get("sample_rate", None)

    def clone(cls):
        """return a deepcopy of self"""
        return copy.deepcopy(cls)

    def __str__(self) -> str:
        """return a descriptive string"""
        info_string = (
            f"Window of {self.num_samples_window} samples with "
            f"overlap {self.num_samples_overlap}"
        )
        return info_string

    def __repr__(self):
        return self.__str__()

    @property
    def num_samples_advance(self) -> int:
        """
        Returns the number of samples the window advances at each step.

        Development Note:
        num_samples_advance is a derived property.  If it were a fundamental
        property then overlap would become a derived property.
        """
        return self.num_samples_window - self.num_samples_overlap

    def available_number_of_windows(self, num_samples_data: int) -> int:
        """
        Returns the number of windows for a dataset with num_samples_data.

        Development Note:
        Only take as many windows as available without wrapping.  Start with one
        window for free, move forward by num_samples_advance and don't walk over
        the cliff.

        Parameters
        ----------
        num_samples_data : int
            The number of samples in the time series to be windowed by self.

        Returns
        -------
        number_of_windows : int
           Count of the number of windows returned from time series of
           num_samples_data.
        """
        return available_number_of_windows_in_array(
            num_samples_data, self.num_samples_window, self.num_samples_advance
        )

    def apply_sliding_window(
        self,
        data: Union[np.ndarray, xr.DataArray, xr.Dataset],
        time_vector: Optional[Union[np.ndarray, None]] = None,
        dt: Optional[Union[float, None]] = None,
        return_xarray: Optional[bool] = False,
    ):
        """
        Applies the windowing scheme (self) to the input data.

        Parameters
        ----------
        data: 1D numpy array, xr.DataArray, xr.Dataset
            The data to break into ensembles.
        time_vector: 1D numpy array
            The time axis of the data.
        dt: float
            The sample interval of the data (reciprocal of sample_rate)
        return_xarray: boolean
            If True will return an xarray object, even if the input object was a
            numpy array

        Returns
        -------
        windowed_obj: arraylike
            Normally an object of type xarray.core.dataarray.DataArray
            Could be numpy array as well.
        """
        if isinstance(data, np.ndarray):
            windowed_obj = self._apply_sliding_window_numpy(
                data, time_vector=time_vector, dt=dt, return_xarray=return_xarray
            )

        elif isinstance(data, xr.DataArray):
            # Cast DataArray to DataSet, iterate and then Dataset back to DataArray
            xrds = data.to_dataset("channel")
            windowed_obj = self.apply_sliding_window(
                xrds, time_vector=time_vector, dt=dt
            )
            windowed_obj = windowed_obj.to_array("channel")

        elif isinstance(data, xr.Dataset):
            ds = xr.Dataset()
            for key in data.keys():
                windowed_obj = self._apply_sliding_window_numpy(
                    data[key].data,
                    time_vector=data.time.data,
                    dt=dt,
                    return_xarray=True,
                )
                ds.update({key: windowed_obj})
            windowed_obj = ds

        else:
            logger.error(f"Unexpected Data type {type(data)}")
            raise Exception
        return windowed_obj

    def _apply_sliding_window_numpy(
        self,
        data: np.ndarray,
        time_vector: Optional[Union[np.ndarray, None]] = None,
        dt: Optional[Union[float, None]] = None,
        return_xarray: Optional[bool] = False,
    ):
        """
        Applies windowing scheme (self) to a numpy array.

        Parameters
        ----------
        data: numpy.ndarray
            A channel of time series data
        time_vector: numpy.ndarray or None
            Time coordinate of xarray.  If None is passed we just assign integer counts
        dt: float or None
            Sampling interval
        return_xarray: bool
            If True an xarray is returned,
            If False we just return a numpy array of the windowed data


        Returns
        -------
        output: xr.DataArray or np.ndarray
            The windowed time series, bound to time axis or just as numpy array,
            depending on the value of return_xarray
        """
        sliding_window_function = SLIDING_WINDOW_FUNCTIONS[self.striding_function_label]
        windowed_array = sliding_window_function(
            data, self.num_samples_window, self.num_samples_advance
        )

        if return_xarray:
            # Get window_time_axis coordinate
            if time_vector is None:
                msg = "xarray requested but time vector not passed -- use integer time axis"
                logger.warning(msg)
                time_vector = np.arange(len(data))
            window_time_axis = self.downsample_time_axis(time_vector)

            output = self.cast_windowed_data_to_xarray(
                windowed_array, window_time_axis, dt=dt
            )
        else:
            output = windowed_array

        return output

    def cast_windowed_data_to_xarray(
        self,
        windowed_array: np.ndarray,
        time_vector: np.ndarray,
        dt: Optional[Union[float, None]] = None,
    ) -> xr.DataArray:
        """
        Casts numpy array to xarray for windowed time series.

        Parameters
        ----------
        windowed_array
        time_vector
        dt

        Returns
        -------
        xr.DataArray:
            Input data with a time and  "within-window time" axis.
        """
        # Get within-window_time_axis coordinate
        if dt is None:
            logger.warning("Warning dt not defined, using dt=1")
            dt = 1.0
        within_window_time_axis = dt * np.arange(self.num_samples_window)

        # cast to xr.DataArray
        xrda = xr.DataArray(
            windowed_array,
            dims=["time", "within-window time"],
            coords={"within-window time": within_window_time_axis, "time": time_vector},
        )
        return xrda

    def left_hand_window_edge_indices(self, num_samples_data: int) -> np.ndarray:
        """Makes an array with the indices of the first sample of each window"""
        if self._left_hand_window_edge_indices is None:
            number_of_windows = self.available_number_of_windows(num_samples_data)
            self._left_hand_window_edge_indices = (
                np.arange(number_of_windows) * self.num_samples_advance
            )
        return self._left_hand_window_edge_indices

    def downsample_time_axis(self, time_axis: np.ndarray) -> np.ndarray:
        """
        Returns a time-axis for the windowed data.

        TODO: Add an option to use window center, instead of forcing LHWE.

        Notes:
        Say that we had 1Hz data starting at t=0 and 100 samples.  Then window,
        with window length 10, and advance 10.  The window_time_axis is
        [0, 10, 20 , ... 90].  If Same window length, but advance were 5.
        Then return [0, 5, 10, 15, ... 90].

        Parameters
        ----------
        time_axis : arraylike
            This is the time axis associated with the time-series prior to
            the windowing operation.

        Returns
        -------
        window_time_axis : array-like
            This is a time axis for the windowed data.  One value per window.


        """
        lhwe = self.left_hand_window_edge_indices(len(time_axis))
        window_time_axis = time_axis[lhwe]
        return window_time_axis

    def apply_taper(self, data):
        """
        modifies the data in place by applying a taper to each window
        """
        data = WindowedTimeSeries.apply_taper(data=data, taper=self.taper)
        return data

    def frequency_axis(self, dt):
        fft_harmonics = get_fft_harmonics(self.num_samples_window, 1.0 / dt)
        return fft_harmonics

    def apply_fft(
        self,
        data: Union[xr.DataArray, xr.Dataset],
        detrend_type: Optional[str] = "linear",
    ) -> xr.Dataset:
        """
        Applies the Fourier transform to each window in the windowed time series.

        Assumes sliding window and taper already applied.

        TODO: Make this return a Spectrogram() object.

        Parameters
        ----------
        data: xarray.core.dataset.Dataset
            The windowed data to FFT
        detrend_type: string
            Passed through to scipy.signal during detrend operation.

        Returns
        -------
        spectral_ds:xr.Dataset
            Dataset same channels as input but data are now complex values Fourier coefficients.

        """
        spectral_ds = WindowedTimeSeries.apply_fft(
            data=data,
            sample_rate=self.sample_rate,
            spectral_density_correction=self.linear_spectral_density_calibration_factor,
            detrend_type=detrend_type,
        )

        return spectral_ds

    # PROPERTIES THAT NEED SAMPLING RATE
    # these may be moved elsewhere later
    @property
    def dt(self) -> float:
        """
        Returns the sample interval of of the time series.
        """
        return 1.0 / self.sample_rate

    @property
    def window_duration(self) -> float:
        """
        Return the duration of the window.
        - Units are those od self.dt (normally seconds)
        """
        return self.num_samples_window * self.dt

    @property
    def duration_advance(self):
        """Return the duration of the window advance"""
        return self.num_samples_advance * self.dt

    @property
    def linear_spectral_density_calibration_factor(self) -> float:
        """
        Gets the calibration factor for Spectral density.

        The factor is applied via multiplication.

        scale_factor = self.linear_spectral_density_calibration_factor
        linear_spectral_data = data * scale_factor

        Following Hienzel et al. 2002, Equations 24 and 25 for Linear Spectral Density
         correction for a single-sided spectrum.

        Returns
        -------
        float
            calibration_factor: Following Hienzel et al 2002,

        """
        return np.sqrt(2 / (self.sample_rate * self.S2))


def window_scheme_from_decimation(decimation: AuroraDecimationLevel):
    """
    Helper function to workaround mt_metadata to not import form aurora

    TODO: Make this a method of AuroraDecimationLevel that returns a WindowingScheme.
     - This requires making WindowingScheme available in mt_metadata.

    Parameters
    ----------
    decimation: AuroraDecimationLevel
        Decimation level metadata object

    Returns
    -------
        windowing_scheme aurora.time_series.windowing_scheme.WindowingScheme
    """
    from aurora.time_series.windowing_scheme import WindowingScheme

    windowing_scheme = WindowingScheme(
        taper_family=decimation.stft.window.type,
        num_samples_window=decimation.stft.window.num_samples,
        num_samples_overlap=decimation.stft.window.overlap,
        taper_additional_args=decimation.stft.window.additional_args,
        sample_rate=decimation.decimation.sample_rate,
    )
    return windowing_scheme
